# Copyright 2020 Brad Schuler K0BAS <brad@schuler.ws>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import struct
import logging
import string

from chirp import bitwise
from chirp import chirp_common
from chirp import directory
from chirp import errors
from chirp import memmap
from chirp import util
from chirp.settings import RadioSetting, RadioSettingGroup, \
    RadioSettingValueInteger, RadioSettingValueList, \
    RadioSettingValueBoolean, RadioSettingValueString, \
    RadioSettingValueFloat, RadioSettings


class ATBankModel(chirp_common.BankModel):
    """AnyTone Banks A-J, Each chan in zero or one bank"""

    def __init__(self, radio, name='Banks'):
        super(ATBankModel, self).__init__(radio, name)
        self._banks = []
        self._memBounds = list(range(0, 750))
        for i in range(0, 10):
            self._banks.append(
                chirp_common.Bank(self, i, string.ascii_uppercase[i]))

    def get_num_mappings(self):
        return len(self._banks)

    def get_mappings(self):
        return self._banks

    def add_memory_to_mapping(self, memory, bank):
        self._radio.set_bank(memory.number, bank.get_index())

    def remove_memory_from_mapping(self, memory, mapping):
        # I would argue that removing a memory from a mapping in which it does
        # not exist should not throw an error.  The end result is the
        # requested result.  The memory is not in the mapping.
        index = self._radio.get_bank(memory.number)
        if (index is None or index != mapping.get_index()):
            raise Exception("Memory %d is not in bank %s" %
                            (memory.number, bank.get_name()))
        self._radio.clr_bank(memory.number)

    def get_mapping_memories(self, bank):
        memories = []
        for i in self._memBounds:
            bank_index = self._radio.get_bank(i)
            if (bank_index is not None and bank_index == bank.get_index()):
                memories.append(self._radio.get_memory(i))
        return memories

    def get_memory_mappings(self, memory):
        bank_index = self._radio.get_bank(memory.number)
        if bank_index is None:
            return []
        else:
            return [self._banks[bank_index]]


LOG = logging.getLogger(__name__)

# TODO: Reverse engineer scramble codes
#       Examples: 1=x0300 1111=xDA0B 3300=x3323 3400=x4524
# TODO: Reverse engineer derived-value for 2Tone frequencies
#       Below are some samples:
#       freq    tone    derived
#               0000    0000
#       288.0   0b40    0823    208.3
#       321.7   0c91    0749    186.5
#       526.8   1494    2397    911.1
#       526.9   1495    2395    910.9
#       712.3   1bd3    1a52    673.8
#       928.1   2441    0286     64.6
#       1527.3  3ba9    0c46    314.2
#       1934.7  4b93    09b1    248.1
#       3116.0  79b8    000c      1.2


mem_format_thruflags = """
struct flag {                   // Flags about a channel
  u8 left_only:1,    // Channel is 1.25m or AM (meaning left side only)
     unused:1,          // Unused flag
     scan:2,            // chirp_common.SKIP_VALUES
     group:4;           // [A-J, Off]
};

#seekto 0x0020;                 // Read Only
struct {
  u8 unknown1[16];
  u8 unknown2[16];
  char modellike[14];
  u8 unknown3[2];
  char serial[16];
  u8 zeros16[16];
  char date[16];
  u8 zeros128[128];
} oem_info;

struct flag chan_flags[750]; // Channel flags
struct flag limit_flags[10]; // Limit channel flags
//u8 effs8[8];
"""

mem_format = mem_format_thruflags + """
struct memory {
    bbcd freq[5];           // rx freq
    bbcd offset[3];         // tx offset from rx / 100
    u8 unknown1:4,
       tune_step:4;         // [2.5k, 5k, 6.25k, 8.33k, 10k,
                            //  12.5k, 15k, 20k, 25k, 30k]
    u8 rxdcsextra:1,        // use with rxcode for index of rx DCS to use
       txdcsextra:1,        // use with txcode for index of tx DCS to use
       rxdcsinv:1,          // inverse DCS rx polarity
       txdcsinv:1,          // inverse DCS tx polarity
       channel_width:2,     // [25 kHz, 20 kHz, 12.5 kHz]
       rev:1,               // reverse the tx and rx frequencies & tones
       txoff:1;             // prevent tx
    u8 talkaround:1,        // Use the rx freq & tone when transmitting
       compander:1,
       unknown2:1,
       is_am:1,             // modulation is AM
       power:2,             // [High, Mid1, Mid2, Low]
       duplex:2;            // ['', -, +, off]
    u8 dtmf_enc_num:4,      // zero-based DTMF slot to use for optional
                            // signalling (when optsig = DTMF)
       rxtmode:2,           // ['', Tone, DTCS]
       txtmode:2;           // ['', Tone, DTCS]
    u8 unknown3:2,
       txtone:6;            // TONES[0..50] & 51=custom
    u8 unknown4:2,
       rxtone:6;            // TONES[0..50] & 51=custom
    u8 txcode;              // use with txdcsextra for index of tx DCS to use
    u8 rxcode;              // use with rxdcsextra for index of rx DCS to use
    u8 unknown5:2,
       pttid:2,             // [Off, Begin, End, Begin&End]
       unknown6:2,
       bclo:2;              // [Off, Repeater, Busy]
    u8 unknown7:6,
       band:2;              // [2m, 1-1/4m, 350+ MHz, 70cm]
    u8 unknown8:5,
       sql_mode:3;          // [Carrier, CTCSS/DCS Tones, Opt Sig Only,
                            //  Tones & Opt Sig, Tones or Opt Sig]
                            // Implemented in GUI as a combination of rxmode
                            // and opt sig squelch list OPT_SIG_SQL
    u8 unknown9:6,
       optsig:2;            // [Off, DTMF, 2Tone, 5Tone]
    u8 unknown10:3,
       twotone:5;           // zero-based 2-tone slot to use for optional
                            // signalling (when optsig = 2Tone)
    u8 unknown11:1,
       fivetone:7;          // zero-based 5-tone slot to use for optional
                            // signalling (when optsig = 5Tone)
    u8 unknown12:4,
       scramble:4;          // [Off, 1 ... 9, Define 1, Define 2]
    char name[7];
    ul16 custtone;          // custom ctcss tone (%5.1f) * 10
};

struct DTMF12p3 {
  u8 dtmf[12];              // Each nibble is a DTMF code [0-9A-D*#]
  u8 num_tones;             // Quantity of nibbles/digits in above
  u8 unknown3[3];
};

struct DTMF7 {
  u8 dtmf[7];               // Each nibble is a DTMF code [0-9A-D*#]
  u8 num_tones;             // Quantity of nibbles/digits in above
};

#seekto 0x0400;                 // DTMF slots
struct DTMF12p3 dtmf_encodings[16];

//#seekto 0x0500;
//u8 unknown[64];

#seekto 0x0540;                 // Hyper channels (12 x 2)
struct memory hyper_channels[24];

struct {
  u8 unknown1:6,
     display:2;             // Display Mode: [freq, chan, name]
  u8 unknown2:3,
     sql_left:5;            // Left Squelch Level: 1..20
  u8 unknown3:3,
     sql_right:5;           // Right Squelch Level: 1..20
  u8 unknown4;
  u8 unknown5:6,
     scan_tot:2;            // Scan Pause Time: [TO, CO, SE]
  u8 unknown6:7,
     scan_mode:1;           // Scan Mode: [MEM, MSM(Pri Scan)]
  u8 bg_red;                // Background color red intensity
                            // 0-x1f displayed as 1-32d
  u8 bg_green;              // Background color green intensity
  u8 bg_blue;               // Background color blue intensity
  u8 unknown7:5,
     tbst_freq:3;           // TBST Freq [Off, 1750, 2100, 1000, 1450] Hz
  u8 unknown8:3,
     ptt_tot:5;             // Talk Timeout Timer: [Off, 1..30] minutes
  u8 unknown9:6,
     ptt_lockout:2;         // PTT Key Lock: [Off, Right, Left, Both]
  u8 unknown10:3,
     apo:5;                 // Auto Power Off: [Off, 1..18] * 30 minutes
  u8 unknown11:3,
     key_bright:5;          // Key Light Level: [00-1e] := [1-32]
  u8 long_key_time;         // Long Key Time: # seconds * 100,
                            // so 100 := 1.0 sec, 250 := 2.5 sec
  u8 unknown12:6,
     deputy_mute:2;         // Mute the Deputy channel (sub band) when
                            // primary is doing this: [Off, Tx, Rx, Tx & Rx]
  //#seekto 0x0850;
  u8 disp_chan_lock:1,      // Display Channel Lock when display=1(Chan)
     unknown13:1,
     keypad_lock:1,         // Lock the keypad
     beep_hi:1,             // 0=low 1=high volume
     no_init:1,             // Prohibit initialization: 0=no 1=yes
     unknown14:3;
  u8 unknown15:2,
     clk_shift:1,           // CLK Shift: 0=off 1=on
     ext_spk_on:1,          // Enable the external speaker
     alt_key_mode:1,        // Use Alt Keypad Mode: 0=off 1=on
     beep_on:1,             // Enable beep
     no_tone_elim_tail:1,   // Elim squelch tail when no CTCSS/DCS signaling
     sql_key_mode:1;        // SQL Key Function: [Momentary, Toggle]
  u8 unknown16:5,
     dtmf_preload_time:3;   // DTMF encode preload time:
                            // [100, 300, 500, 800, 1000] ms
  u8 unknown17:5,
     dtmf_speed:3;          // DTMF transmit time/speed:
                            // [30, 50, 80, 100, 150, 200, 250] ms
  u8 unknown19:6,
     tail_elim_type:2;      // Tail Eliminator Type: [Off, 120, 180] degrees
  u8 unknown20:3,
     key_a_func:5;          // Programmable key A function [0-19]:
                            // [Off, DUP, ... Bandwidth, Talk Around]
  u8 unknown21:3,
     key_b_func:5;          // Programmable key B function
  u8 unknown22:3,
     key_c_func:5;          // Programmable key C function
  u8 unknown23:3,
     key_d_func:5;          // Programmable key D function
  u8 unknown24:7,
     boot_pw_on:1;          // Use boot password (verify password is set!)
  u8 unknown25:5,
     rfsql_left:3;          // Left RF Squelch Level:
                            // [Off, S-2, S-5, S-9, FULL]
  u8 unknown26:5,
     rfsql_right:3;         // Right RF Squelch Level
  u8 scramble1[2];          // Scramble code #1 (unknown encoding)
  u8 scramble2[2];          // Scramble code #2 (unknown encoding)

  //#seekto 0x0860 & 0x0880
  struct {
    u8 unknown1;
    u8 unknown2:6,
       left_mode:2;         // [VFO, Channel, Home]
    u8 unknown3:6,
       right_mode:2;        // [VFO, Channel, Home]
    u8 unknown4:6,
       sub_display:2;       // [Freq, DC IN, Off]
    ul16 left_channel;      // Zero-based chan number
    ul16 right_channel;     // Zero-based chan number
    u8 unknown5;
    u8 unknown6:6,
       spkr_mode:2;         // Hand/Main [Off/On, On/On, On/Off]
    u8 unknown7:6,
       left_vfo_band:2;     // [108, 220, 350, 400] MHz
    u8 unknown8:7,
       right_vfo_band:1;    // [136, 400] MHz
    u8 unknown9;
    u8 unknown10:1,
       vfo_tracked:1,
       auto_hyper_save:1,
       auto_rpt_shift:1,
       auto_am:1,
       main:1,              // [Left, Right]
       vfo_band_edge:1,
       unknown11:1;
    u8 unknown12;
    u8 unknown13;
    //#seekto 0x0870 & 0x0890
    u8 unknown14:4,
       left_work_bank:4;    // [A - J, Off]
    u8 unknown15:7,
       left_bank_mode:1;    // [CH, Bank]
    u8 unknown16:7,
       left_bank_sw:1;      // [Off, On]
    u8 unknown17:4,
       right_work_bank:4;   // [A - J, Off]
    u8 unknown18:7,
       right_bank_mode:1;   // [CH, Bank]
    u8 unknown19:7,
       right_bank_sw:1;     // [Off, On]
    // This would be better than below:
    //   struct { u8 unknown:7, linked:1; } linkedBanks[10];
    // They would need to be in a sub-group of settings in the extras.
    // A-J (one byte each) 00 or 01
    u8 unknown20:7,
        link_bank_A:1;
    u8 unknown21:7,
        link_bank_B:1;
    u8 unknown22:7,
        link_bank_C:1;
    u8 unknown23:7,
        link_bank_D:1;
    u8 unknown24:7,
        link_bank_E:1;
    u8 unknown25:7,
        link_bank_F:1;
    u8 unknown26:7,
        link_bank_G:1;
    u8 unknown27:7,
        link_bank_H:1;
    u8 unknown28:7,
        link_bank_I:1;
    u8 unknown29:7,
        link_bank_J:1;
  } hyper_settings[2];

  //#seekto 0x08a0;
  char boot_password[7];    // Boot password
  u8 unknown30[9];

  //#seekto 0x08b0;
  u8 unknown31[16];

  //#seekto 0x08c0;
  u8 unknown32:4,
     dtmf_interval_char:4;  // DTMF interval character
  u8 dtmf_group_code;       // DTMF Group Code: 0a-0f,ff displayed as A-#,Off
  u8 unknown33:6,
     dtmf_dec_resp:2;       // DTMF Decoding Response:
                            // [None, Beep, Beep & Respond]
  u8 unknown34;
  u8 dtmf_first_dig_time;   // DTMF First digit time: 0..250 * 10 ms
  u8 dtmf_auto_rst_time;    // DTMF auto reset time: 0..250
                            // displayed as 0.0-25.0 in 0.1 increments sec
  u8 unknown35;
  u8 dtmf_self_id[3];       // DTMF Self ID: low nibble of each byte is a
                            // DTMF char, exactly 3 chars
  u8 unknown36;
  u8 unknown37;
  u8 unknown38;
  u8 unknown39:7,
     dtmf_sidetone_on:1;    // DTMF side tone enabled
  u8 dtmf_enc_delay;        // DTMF encode delay: 1..250 * 10 ms
  u8 dtmf_pttid_delay;      // DTMF PTT ID delay: 0,5-75
                            // shown as [Off,5-75] sec

  //#seekto 0x08d0;
  struct DTMF7 dtmf_kill;    // DTMF to remotely kill
  struct DTMF7 dtmf_stun;    // DTMF to remotely stun
} settings;

//#seekto 0x08e0;                   // 2 Tone settings - for future work
//struct {
//  ul16 calltone1;           // Call format 1st tone freq * 10
//  ul16 calltone2;           // Call format 2nd tone freq * 10
//                            // (or 00 00 for long tone version of 1st tone)
//  ul16 calltone1derived;    // some derived freq? calculated from 1st tone
//  ul16 calltone2derived;    // some derived freq? calculated from 2nd tone
//  u8 unknown1:6,
//     decode_resp:2;         // Decode response [None, Beep, Beep & Respond]
//  u8 tone1_dur;             // 1st tone duration * 10
//                            // shown as range(0.5, 10, 0.1) seconds
//  u8 tone2_dur;             // 2nd tone duration * 10
//                            // shown as range(0.5, 10, 0.1) seconds
//  u8 longtone_dur;          // Long tone duration * 10
//                            // shown as range(0.5, 10, 0.1) seconds
//  u8 gap_time;              // Gap time / 100
//                            // shown as range(0, 2000, 100) msec
//  u8 auto_rst_time;         // Auto reset time * 10
//                            // shown as range(0, 25, 0.1) sec
//  u8 unknown:7,
//     enc_sidetone;          // Encode side-tone 0=off 1=on
//  u8 unknown2:4,
//     call_format:4;         // Call format [A-B, A-C, A-D, B-A, B-C, B-D,
//                            //              C-A, C-B, C-D, D-A, D-B, D-C,
//                            //              Long A, Long B, Long C]
//                            // Be sure to change callTone1/2 and
//                            // callTone1/2derived when callFormat is changed
//  //#seekto 0x08f0;
//  u16 a_tone;               // A tone freq * 10
//  u16 b_tone;               // B tone freq * 10
//  u16 c_tone;               // C tone freq * 10
//  u16 d_tone;               // D tone freq * 10
//  u8 zeros[8];
//} twotone_settings;

//#seekto 0x0900;                 // Unknown all 0x00
//u8 zeros[512];

//#seekto 0x0b00;                 // 2 Tone memories (24 slots of 16 bytes)
//struct {
//  ul16 tone1derived;        // some derived freq? calculated from 1st tone
//  ul16 tone2derived;        // some derived freq? calculated from 2nd tone
//  ul16 tone1;               // 1st tone freq * 10
//  ul16 tone2;               // 2nd tone freq * 10
//  char name[7];             // Tone name (padded)
//  u8 zero;
//} twotone_slots[24];

//#seekto 0x0c80;                 // Unknown all 0x24
//u8 twentyfours[128];

//#seekto 0x0d00;                 // 5 Tone configs (100 slots of 32 bytes)
//u8 fivetones[3200];

#seekto 0x1980;                 // Communication  notes
struct {
    char call_id[5];         // Call ID (numeric, max 5 digits (x30-x39)
    u8 zeros[3];            // 0x00 * 3
} note_call_ids[100];
struct {
    char name[7];           // Names (ALPHAnumeric max 7 chars)
    char space;             // 0x20
} note_names[100];

//#seekto 0x1fc0;                 // Unknown all 0x00
//u8 zeros[64];

#seekto 0x2000;
struct memory channels[750];        // Normal channels (750 * 32 bytes)

//#seekto 0x7dc0;
struct memory limit_channels[10];   // Limit channels
                                    // (10 * 32 bytes: Five pair of lo/hi)

//#seekto 0x7f00;                 // Unknown
//u8 unknown[32];

#seekto 0x7f20;
struct DTMF12p3 dtmf_pttid_bot;   // DTMF PTT ID Start (BOT)
struct DTMF12p3 dtmf_pttid_eot;   // DTMF PTT ID End (EOT)

//#seekto 0x7f40;
//u8 unknown[64]

#seekto 0x7f80;                 // Emergency Information
struct {
  u8 unknown:6,
     mode:2;                // [Alarm, Transpond+Background,
                            //  Transpond+Alarm, Both]
  u8 unknown2:6,
     eni_type:2;            // [None, DTMF, 5Tone]
  u8 id;                    // When DTMF: [M1-M16], When 5Tone: 0-99
  u8 alarm_time;            // [1-255] seconds
  u8 tx_time;               // [0-255] seconds
  u8 rx_time;               // [0-255] seconds
  u8 unknown3:7,
     chan_select:1;         // [Assigned, Selected]
  ul16 channel;             // [0-749]
  u8 cycle;                 // [Continuous, 1-255]
} emergency;
u8 unknown9[6];

//#seekto 0x7f90;                 // Unknown
//u8 unknown[48]

#seekto 0x7fc0;                 // Welcome message
char welcome[7];
//u8 zeros[9];

//#seekto 0x7fd0;
//u8 unknown[48];
"""


def _is_chan_used(memobj, num):
    return not memobj.chan_flags[num].unused


def _is_limit_chan_used(memobj, num):
    return not memobj.limit_flags[num].unused


def _should_send_addr(memobj, addr):
    # Skip read-only & out of bounds
    if addr < 0x0100 or addr >= 0x8000:
        return False
    # Skip regions of unknown or uncontrolled by this software
    if ((addr >= 0x0500 and addr < 0x0540) or
            (addr >= 0x08e0 and addr < 0x1980) or
            (addr >= 0x1fc0 and addr < 0x2000) or
            (addr >= 0x7f00 and addr < 0x7f20) or
            (addr >= 0x7f40 and addr < 0x7f80) or
            (addr >= 0x7f90 and addr < 0x7fc0) or
            addr >= 0x7fd0):
        return False
    # Skip unused memories
    if addr >= 0x2000 and addr < 0x7dc0:
        return _is_chan_used(memobj, int((addr - 0x2000) / 0x20))
    if addr >= 0x7dc0 and addr < 0x7f00:
        return _is_limit_chan_used(memobj, int((addr - 0x7dc0) / 0x20))
    return True


def _echo_write(radio, data):
    try:
        radio.pipe.write(data)
        radio.pipe.read(len(data))
    except Exception as e:
        LOG.error("Error writing to radio: %s" % e)
        raise errors.RadioError("Unable to write to radio")


def _read(radio, length):
    try:
        data = radio.pipe.read(length)
    except Exception as e:
        LOG.error("Error reading from radio: %s" % e)
        raise errors.RadioError("Unable to read from radio")

    if len(data) != length:
        LOG.error("Short read from radio (%i, expected %i)" %
                  (len(data), length))
        LOG.debug(util.hexprint(data))
        raise errors.RadioError("Short read from radio")
    return data


valid_model = [b'I588UVP']


def _ident(radio):
    radio.pipe.timeout = 1
    _echo_write(radio, b"PROGRAM")
    response = radio.pipe.read(3)
    if response != b"QX\x06":
        LOG.debug("Response was:\n%s" % util.hexprint(response))
        raise errors.RadioError("Unsupported model")
    _echo_write(radio, b"\x02")
    response = radio.pipe.read(16)
    LOG.debug(util.hexprint(response))
    if response[-1:] != b"\x06":
        LOG.debug("Response was:\n%s" % util.hexprint(response))
        raise errors.RadioError("Missing ack")
    if response[0:7] not in valid_model:
        LOG.debug("Response was:\n%s" % util.hexprint(response))
        raise errors.RadioError("Unsupported model")
    # Manufacturer software does this also: _send(radio, b'R', 0x0080, 0x10)


def _finish(radio):
    endframe = b"\x45\x4E\x44"
    _echo_write(radio, endframe)
    result = radio.pipe.read(1)
    if result != b"\x06":
        LOG.debug("Got:\n%s" % util.hexprint(result))
        raise errors.RadioError("Radio did not finish cleanly")


def _checksum(data):
    cs = 0
    for byte in data:
        cs += byte
    return cs % 256


def _send(radio, cmd, addr, length, data=None):
    frame = struct.pack(">cHb", cmd, addr, length)
    if data:
        frame += data
        frame += bytes([_checksum(frame[1:])])
        frame += b"\x06"
    _echo_write(radio, frame)
    LOG.debug("Sent:\n%s" % util.hexprint(frame))
    if data:
        result = radio.pipe.read(1)
        if result != b"\x06":
            LOG.debug("Ack was: %s" % util.hexprint(result))
            raise errors.RadioError(
                "Radio did not accept block at %04x" % addr)
        return
    result = _read(radio, length + 6)
    LOG.debug("Got:\n%s" % util.hexprint(result))
    header = result[0:4]
    data = result[4:-2]
    ack = result[-1:]
    if ack != b"\x06":
        LOG.debug("Ack was: %s" % repr(ack))
        raise errors.RadioError("Radio NAK'd block at %04x" % addr)
    _cmd, _addr, _length = struct.unpack(">cHb", header)
    if _addr != addr or _length != _length:
        LOG.debug("Expected/Received:")
        LOG.debug(" Length: %02x/%02x" % (length, _length))
        LOG.debug(" Addr: %04x/%04x" % (addr, _addr))
        raise errors.RadioError("Radio send unexpected block")
    cs = _checksum(result[1:-2])
    if cs != result[-2]:
        LOG.debug("Calculated: %02x" % cs)
        LOG.debug("Actual:     %02x" % result[-2])
        raise errors.RadioError("Block at 0x%04x failed checksum" % addr)
    return data


def _download(radio):
    _ident(radio)

    memobj = None

    data = b""
    for start, end in radio._ranges:
        for addr in range(start, end, 0x10):
            if memobj is not None and not _should_send_addr(memobj, addr):
                block = b"\x00" * 0x10
            else:
                block = _send(radio, b'R', addr, 0x10)
            data += block

            status = chirp_common.Status()
            status.cur = len(data)
            status.max = end
            status.msg = "Cloning from radio"
            radio.status_fn(status)

            if addr == 0x0400 - 0x10:
                memobj = bitwise.parse(mem_format_thruflags, data)

    _finish(radio)

    return memmap.MemoryMapBytes(data)


def _upload(radio):
    _ident(radio)

    for start, end in radio._ranges:
        for addr in range(start, end, 0x10):
            if not _should_send_addr(radio._memobj, addr):
                continue
            block = radio._mmap[addr:addr + 0x10]
            _send(radio, b'W', addr, len(block), block)

            status = chirp_common.Status()
            status.cur = addr
            status.max = end
            status.msg = "Cloning to radio"
            radio.status_fn(status)

    _finish(radio)


def _filter(s, charset):
    rc = ""
    for i in range(0, len(s)):
        c = str(s[i])
        rc += (c if c in charset else "")
    return rc


SEVEN_SPACES = "       "
APO = ['Off'] + ['%.1f hour(s)' % (0.5 * x) for x in range(1, 25)]
ONEANDQTRBAND = (220000000, 260000000)
LEFT_BANDS = [
    (108000000, 180000000),  # L-108
    ONEANDQTRBAND,           # L-220
    (350000000, 399995000),  # L-350
    (400000000, 523000000)   # L/R-400
]
RIGHT_BANDS = [
    (136000000, 174000000),  # R-136
    (400000000, 523000000)   # L/R-400
]
BCLO = ['Off', 'Repeater', 'Busy']
BEEP_VOL = ['Low', 'High']
CROSS_MODES = list(chirp_common.CROSS_MODES)
CROSS_MODES.remove("Tone->")
DISPLAY = ['Freq', 'Chan#', 'Name']
DTMF_SLOTS = ['M%d' % x for x in range(1, 17)]
DUPLEXES = ['', '-', '+', 'off']
KEY_FUNCS = ['Off', 'DUP', 'PRI On', 'Power', 'Set DCS/CTCSS Code', 'MHz',
             'Reverse', 'HM Chan', 'Main L/R Switch', 'VFO/MR', 'Scan',
             'Squelch Off', 'Call TBST', 'Call', 'Tone Compande',
             'Scramble', 'Add Opt Signal', 'Bandwidth', 'Talk Around']
MODES = ["WFM", "FM", "NFM"]
OPT_SIGS = ['Off', 'DTMF', '2Tone', '5Tone']
POWER_LEVELS = [chirp_common.PowerLevel("High", watts=50),
                chirp_common.PowerLevel("Mid1", watts=25),
                chirp_common.PowerLevel("Mid2", watts=10),
                chirp_common.PowerLevel("Low", watts=5)]
PTT_IDS = ['Off', 'Begin', 'End', 'Begin & End']
RF_SQUELCHES = ['Off', 'S-2', 'S-5', 'S-9', 'FULL']
SCAN_PAUSES = ['TO', 'CO', 'SE']
SCAN_MODES = ['MEM', 'MSM (Pri Scan)']

SCRAMBLE_CODES = ['Off']
for x in range(1, 10):
    SCRAMBLE_CODES.append('%d' % x)
SCRAMBLE_CODES += ['Define 1', 'Define 2']

T_MODES = ['', 'Tone', 'DTCS', '']
TONE2_SLOTS = ['%d' % x for x in range(0, 24)]
TONE5_SLOTS = ['%d' % x for x in range(0, 100)]
# Future: chirp_common should present TONES instead of hard-coded list
TONES = (62.5,) + chirp_common.TONES
TOT = ['Off'] + ['%s Minutes' % x for x in range(1, 31)]
TUNING_STEPS = [2.5, 5, 6.25, 8.33, 10, 12.5, 15, 20, 25, 30]
PTT_KEY_LOCKS = ["Off", "Right", "Left", "Both"]
TBST_FREQS = ["Off", "1750 Hz", "2100 Hz", "1000 Hz", "1450 Hz"]
LONG_KEY_TIMES = ["1 second", "1.5 seconds", "2 seconds", "2.5 seconds"]
DEPUTY_CHAN_MUTES = ["Off", "Tx", "Rx", "Both"]
SQL_BTN_MODES = ["Momentary", "Toggle"]
SQL_MODES = ["Carrier", "CTCSS/DCS", "Opt Sig Only", "Tones AND Sig",
             "Tones OR Sig"]
OPT_SIG_SQL = ["Off"] + SQL_MODES[2:]
TAIL_ELIM_TYPES = ["Off", "120 degrees", "180 degrees"]
LIMIT_NAMES = ["Limit %s %d" % (c, i + 1)
               for i in range(0, 5) for c in ('Lo', 'Hi')]
HYPER_NAMES = ["Hyper-%s %s%s" % (h, x, y)
               for h in ('1', '2')
               for x in ('', 'H')
               for y in ('L-108', 'L-220', 'L-350', 'L-400', 'R-136', 'R-400')]
HYPER_MODES = ['VFO', 'Channel', 'Home']
HYPER_SUB_DISPLAYS = ['Freq', 'DC IN', 'Off']
HYPER_SPKR_MODES = ['Mic off', 'Mic on', 'Main off']
HYPER_L_VFOS = ['108 MHz', '220 MHz', '350 MHz', '400 MHz']
HYPER_R_VFOS = ['136 MHz', '400 MHz']
HYPER_MAINS = ['Left', 'Right']
HYPER_BANK_MODES = ['CH', 'Bank']
EMER_MODES = ['Alarm', 'Transpond+Background', 'Transpond+Alarm', 'Both']
EMER_ENI_TYPES = ['None', 'DTMF', '5Tone']
EMER_CHAN_SEL = ['Assigned', 'Selected']
EMER_CYCLES = ['Continuous'] + ["%d seconds" % x for x in range(1, 256)]
BANK_CHOICES = ["%s" % chr(x) for x in range(ord("A"), ord("J") + 1)] + ['Off']
DTMF_CHARS = "0123456789ABCD*#"
DTMF_INTCHARS = "ABCD*#"
DTMF_PRELOADS = ['100 ms', '300 ms', '500 ms', '800 ms', '1000 ms']
DTMF_SPEEDS = ['30 ms', '50 ms', '80 ms', '100 ms',
               '150 ms', '200 ms', '250 ms']
DTMF_GROUPS = list("ABCDEF") + ['Off']
DTMF_RESPONSES = ['None', 'Beep', 'Beep & Respond']
DTMF_PTTDELAYS = ['Off'] + ["%d seconds" % x for x in range(5, 76)]


def _dtmf_encode(dtmf_str, numBytes):
    dtmf_bytes = []
    if len(dtmf_str) != 0:
        _byte = 0x00
        for i in range(1, len(dtmf_str)+1):
            if i % 2 == 0:
                _byte |= DTMF_CHARS.index(dtmf_str[i-1])
                dtmf_bytes.append(_byte)
                _byte = 0x00
            else:
                _byte = DTMF_CHARS.index(dtmf_str[i-1]) << 4
        if len(dtmf_str) % 2 == 1:
            dtmf_bytes.append(_byte)
    while (len(dtmf_bytes) < numBytes):
        dtmf_bytes.append(0)
    return dtmf_bytes


def _dtmf_decode(dtmf_bytes, num_tones):
    dtmf_str = ""
    x = 1
    while (x <= num_tones):
        _byte = dtmf_bytes[(x-1)/2]
        dtmf_str += DTMF_CHARS[_byte >> 4]
        x += 1
        if (x <= num_tones):
            dtmf_str += DTMF_CHARS[_byte & 0x0F]
            x += 1
    return dtmf_str


class _RadioSettingValueOffsetInt(RadioSettingValueInteger):
    """An integer setting whose real value is offset from displayed value"""

    def __init__(self, minval, maxval, current, step=1, offset=0):
        RadioSettingValueInteger.__init__(self, minval, maxval, current, step)
        self.offset = offset


@directory.register
class AnyTone5888UVIIIRadio(chirp_common.CloneModeRadio,
                            chirp_common.ExperimentalRadio):
    """AnyTone 5888UVIII"""
    VENDOR = "AnyTone"
    MODEL = "5888UVIII"
    BAUD_RATE = 9600
    NEEDS_COMPAT_SERIAL = False
    _file_ident = b"588UVP"

    _ranges = [
            (0x0000, 0x8000)
        ]

    def get_bank_model(self):
        return ATBankModel(self)

    @classmethod
    def get_prompts(cls):
        rp = chirp_common.RadioPrompts()
        rp.experimental = ("The Anytone 5888UVIII driver is currently "
                           "experimental. There are no known issues with it, "
                           "but you should proceed with caution. 2-tone, "
                           "5-tone, and scramble settings are limited. ")
        return rp

    def get_features(self):
        rf = chirp_common.RadioFeatures()
        rf.has_rx_dtcs = True
        rf.has_bank = True
        rf.has_cross = True
        rf.has_settings = True
        rf.valid_modes = MODES + ['AM']
        rf.valid_tmodes = ['', 'Tone', 'TSQL', 'DTCS', 'Cross']
        rf.valid_cross_modes = CROSS_MODES
        rf.valid_duplexes = DUPLEXES
        rf.valid_tuning_steps = TUNING_STEPS
        rf.valid_bands = LEFT_BANDS + RIGHT_BANDS
        rf.valid_skips = chirp_common.SKIP_VALUES
        rf.valid_power_levels = POWER_LEVELS
        rf.valid_characters = chirp_common.CHARSET_UPPER_NUMERIC + "-"
        rf.valid_name_length = 7
        # Future: rf.valid_tones = TONES
        # Future: update memdetail.py to use rf.valid_dtcs_codes
        # Future: update memdetail.py to use rf.dtcs_polarity
        rf.valid_dtcs_codes = chirp_common.ALL_DTCS_CODES
        rf.memory_bounds = (1, 750)
        rf.valid_special_chans = LIMIT_NAMES + HYPER_NAMES
        return rf

    def sync_in(self):
        self._mmap = _download(self)
        self.process_mmap()

    def sync_out(self):
        _upload(self)

    def process_mmap(self):
        self._memobj = bitwise.parse(mem_format, self._mmap)

    def _get_memobjs(self, number):
        if number <= len(self._memobj.channels):
            _mem = self._memobj.channels[number - 1]
            _flg = self._memobj.chan_flags[number - 1]
        elif (number <= len(self._memobj.channels) +
              len(self._memobj.limit_channels)):
            _mem = self._memobj.limit_channels[number -
                                               len(self._memobj.channels) -
                                               1]
            _flg = self._memobj.limit_flags[number -
                                            len(self._memobj.channels) -
                                            1]
        else:
            _mem = self._memobj.hyper_channels[
                number -
                len(self._memobj.channels) -
                len(self._memobj.limit_channels) -
                1]
            _flg = None
        return _mem, _flg

    def _get_dcs_index(self, _mem, which):
        base = getattr(_mem, '%scode' % which)
        extra = getattr(_mem, '%sdcsextra' % which)
        return (int(extra) << 8) | int(base)

    def _set_dcs_index(self, _mem, which, index):
        base = getattr(_mem, '%scode' % which)
        extra = getattr(_mem, '%sdcsextra' % which)
        base.set_value(index & 0xFF)
        extra.set_value(index >> 8)

    def get_raw_memory(self, number):
        _mem, _flg = self._get_memobjs(number)
        return repr(_mem) + repr(_flg)

    def set_bank(self, number, bank_index):
        _mem, _flg = self._get_memobjs(number)
        _flg.group = bank_index

    def clr_bank(self, number):
        _mem, _flg = self._get_memobjs(number)
        _flg.group = BANK_CHOICES.index('Off')

    def get_bank(self, number):
        _mem, _flg = self._get_memobjs(number)
        if (_flg.group < 0xa):
            return _flg.group
        return None

    def get_memory(self, number):
        mem = chirp_common.Memory()
        is_limit_chan = is_hyper_chan = False
        if isinstance(number, str):
            if number in LIMIT_NAMES:
                is_limit_chan = True
                mem.number = (len(self._memobj.channels) +
                              LIMIT_NAMES.index(number) +
                              1)
            elif number in HYPER_NAMES:
                is_hyper_chan = True
                mem.number = (len(self._memobj.channels) +
                              len(LIMIT_NAMES) +
                              HYPER_NAMES.index(number) +
                              1)
            mem.extd_number = number
        elif number > (len(self._memobj.channels) +
                       len(self._memobj.limit_channels)):
            is_hyper_chan = True
            mem.number = number
            mem.extd_number = HYPER_NAMES[number -
                                          len(self._memobj.channels) -
                                          len(self._memobj.limit_channels) -
                                          1]
        elif number > len(self._memobj.channels):
            is_limit_chan = True
            mem.number = number
            mem.extd_number = LIMIT_NAMES[number -
                                          len(self._memobj.channels) -
                                          1]
        else:
            mem.number = number

        _mem, _flg = self._get_memobjs(mem.number)

        if not is_hyper_chan and _flg.unused:
            mem.empty = True
            _mem.fill_raw(b"\x00")
            _mem.name = SEVEN_SPACES
            _flg.scan = 0

        mem.freq = int(_mem.freq)

        mem.offset = int(_mem.offset) * 100
        mem.name = (str(_mem.name).rstrip()
                    if not is_limit_chan and not is_hyper_chan
                    else SEVEN_SPACES)
        mem.duplex = DUPLEXES[_mem.duplex]
        mem.mode = _mem.is_am and "AM" or MODES[_mem.channel_width]
        mem.tuning_step = TUNING_STEPS[_mem.tune_step]

        if _mem.txoff:
            mem.duplex = DUPLEXES[3]

        rxtone = txtone = None
        rxmode = T_MODES[_mem.rxtmode]
        if _mem.sql_mode == 0 or _mem.sql_mode == 2:
            rxmode = T_MODES.index('')
        txmode = T_MODES[_mem.txtmode]
        if txmode == "Tone":
            # If custom tone is being used, show as 88.5 and set checkbox in
            # extras.
            # Future: Improve chirp_common, so I can add "CUSTOM" into TONES
            if _mem.txtone == len(TONES):
                txtone = 88.5
            else:
                txtone = TONES[_mem.txtone]
        elif txmode == "DTCS":
            txtone = chirp_common.ALL_DTCS_CODES[self._get_dcs_index(_mem,
                                                                     'tx')]
        if rxmode == "Tone":
            # If custom tone is being used, show as 88.5 and set checkbox in
            # extras.
            # Future: Improve chirp_common, so I can add "CUSTOM" into TONES
            if _mem.rxtone == len(TONES):
                rxtone = 88.5
            else:
                rxtone = TONES[_mem.rxtone]
        elif rxmode == "DTCS":
            rxtone = chirp_common.ALL_DTCS_CODES[self._get_dcs_index(_mem,
                                                                     'rx')]

        rxpol = _mem.rxdcsinv and "R" or "N"
        txpol = _mem.txdcsinv and "R" or "N"

        chirp_common.split_tone_decode(mem,
                                       (txmode, txtone, txpol),
                                       (rxmode, rxtone, rxpol))

        mem.skip = chirp_common.SKIP_VALUES[_flg.scan
                                            if not is_hyper_chan
                                            else 0]
        mem.power = POWER_LEVELS[_mem.power]

        mem.extra = RadioSettingGroup("Extra", "extra")

        rs = RadioSetting("rev", "Reverse",
                          RadioSettingValueBoolean(_mem.rev))
        mem.extra.append(rs)

        rs = RadioSetting("compander", "Compander",
                          RadioSettingValueBoolean(_mem.compander))
        mem.extra.append(rs)

        rs = RadioSetting("talkaround", "Talkaround",
                          RadioSettingValueBoolean(_mem.talkaround))
        mem.extra.append(rs)

        rs = RadioSetting("pttid", "PTT ID",
                          RadioSettingValueList(PTT_IDS,
                                                PTT_IDS[_mem.pttid]))
        mem.extra.append(rs)

        rs = RadioSetting("bclo", "Busy Channel Lockout",
                          RadioSettingValueList(BCLO,
                                                BCLO[_mem.bclo]))
        mem.extra.append(rs)

        rs = RadioSetting("optsig", "Optional Signaling",
                          RadioSettingValueList(OPT_SIGS,
                                                OPT_SIGS[_mem.optsig]))
        mem.extra.append(rs)

        rs = RadioSetting("OPTSIGSQL", "Squelch w/Opt Signaling",
                          RadioSettingValueList(
                              OPT_SIG_SQL,
                              SQL_MODES[_mem.sql_mode]
                              if SQL_MODES[_mem.sql_mode] in OPT_SIG_SQL
                              else "Off"))
        mem.extra.append(rs)

        rs = RadioSetting("dtmf_enc_num", "DTMF",
                          RadioSettingValueList(DTMF_SLOTS,
                                                DTMF_SLOTS[_mem.dtmf_enc_num]))
        mem.extra.append(rs)

        rs = RadioSetting("twotone", "2-Tone",
                          RadioSettingValueList(TONE2_SLOTS,
                                                TONE2_SLOTS[_mem.twotone]))
        mem.extra.append(rs)

        rs = RadioSetting("fivetone", "5-Tone",
                          RadioSettingValueList(TONE5_SLOTS,
                                                TONE5_SLOTS[_mem.fivetone]))
        mem.extra.append(rs)

        rs = RadioSetting("scramble", "Scrambler Switch",
                          RadioSettingValueList(SCRAMBLE_CODES,
                                                SCRAMBLE_CODES[_mem.scramble]))
        mem.extra.append(rs)

        # Memory properties dialog is only capable of Boolean and List
        # RadioSettingValue classes, so cannot configure custtone.
        # rs = RadioSetting("custtone", "Custom CTCSS",
        #                   RadioSettingValueFloat(
        #                       min(TONES),
        #                       max(TONES),
        #                       _mem.custtone and _mem.custtone / 10 or 151.1,
        #                       0.1,
        #                       1))
        # mem.extra.append(rs)
        custtone_str = chirp_common.format_freq(_mem.custtone)

        rs = RadioSetting("CUSTTONETX",
                          "Use Custom CTCSS (%s) for Tx" % custtone_str,
                          RadioSettingValueBoolean(_mem.txtone == len(TONES)))
        mem.extra.append(rs)

        rs = RadioSetting("CUSTTONERX",
                          "Use Custom CTCSS (%s) for Rx" % custtone_str,
                          RadioSettingValueBoolean(_mem.rxtone == len(TONES)))
        mem.extra.append(rs)

        return mem

    def set_memory(self, mem):
        is_limit_chan = is_hyper_chan = False
        if mem.number > (len(self._memobj.channels) +
                         len(self._memobj.limit_channels)):
            is_hyper_chan = True
        elif mem.number > len(self._memobj.channels):
            is_limit_chan = True

        _mem, _flg = self._get_memobjs(mem.number)

        if mem.empty:
            if is_hyper_chan:
                raise errors.InvalidValueError("Hyper memories may not be "
                                               "empty.")
            _mem.set_raw("\x00" * 32)
            _flg.unused = 1
            _flg.group = BANK_CHOICES.index('Off')
            return

        if mem.mode == "AM" and (mem.freq < LEFT_BANDS[0][0] or
                                 mem.freq > LEFT_BANDS[0][1]):
            raise errors.InvalidValueError("AM modulation only allowed in 2m "
                                           "band.")

        _mem.set_raw("\x00" * 32)
        if _flg:
            # Only non-special memories track this
            _flg.unused = 0

        _mem.freq = mem.freq

        for band_idx in range(0, len(LEFT_BANDS)):
            if (mem.freq >= LEFT_BANDS[band_idx][0] and
                    mem.freq <= LEFT_BANDS[band_idx][1]):
                _mem.band = band_idx
                break

        if _flg:
            # Only non-special memories have this
            _flg.left_only = 0
        if (not is_hyper_chan):
            if mem.mode == "AM":
                _flg.left_only = 1
            else:
                _flg.left_only = 1
                for low, high in RIGHT_BANDS:
                    if mem.freq >= low and mem.freq <= high:
                        _flg.left_only = 0
                        break

        _mem.offset = mem.offset / 100

        if is_hyper_chan:
            mem.name = SEVEN_SPACES
        else:
            _mem.name = mem.name.ljust(7)

        if mem.duplex == "off":
            _mem.duplex = DUPLEXES.index("")
            _mem.txoff = 1
        else:
            _mem.duplex = DUPLEXES.index(mem.duplex)
            _mem.txoff = 0

        _mem.is_am = mem.mode == "AM"

        _mem.tune_step = TUNING_STEPS.index(mem.tuning_step)

        try:
            _mem.channel_width = MODES.index(mem.mode)
        except ValueError:
            _mem.channel_width = 0

        ((txmode, txtone, txpol),
         (rxmode, rxtone, rxpol)) = chirp_common.split_tone_encode(mem)

        _mem.txtmode = T_MODES.index(txmode)
        _mem.rxtmode = T_MODES.index(rxmode)
        if rxmode != '':
            _mem.sql_mode = SQL_MODES.index("CTCSS/DCS")
        else:
            _mem.sql_mode = SQL_MODES.index("Carrier")
        if txmode == "Tone":
            _mem.txtone = TONES.index(txtone)
        elif txmode == "DTCS":
            self._set_dcs_index(_mem, 'tx',
                                chirp_common.ALL_DTCS_CODES.index(txtone))
        if rxmode == "Tone":
            _mem.rxtone = TONES.index(rxtone)
        elif rxmode == "DTCS":
            self._set_dcs_index(_mem, 'rx',
                                chirp_common.ALL_DTCS_CODES.index(rxtone))

        _mem.txdcsinv = txpol == "R"
        _mem.rxdcsinv = rxpol == "R"

        if not is_hyper_chan:
            _flg.scan = chirp_common.SKIP_VALUES.index(mem.skip)

        if mem.power:
            _mem.power = POWER_LEVELS.index(mem.power)
        else:
            _mem.power = 0

        for setting in mem.extra:
            if setting.get_name() == "ignore":
                LOG.debug("*** ignore: %s" % str(setting.value))
            # Future: elif setting.get_name() == "custtone":
            # Future: setattr(_mem, "custtone", setting.value.get_value() * 10)
            elif setting.get_name() == "OPTSIGSQL":
                if str(setting.value) != "Off":
                    _mem.sql_mode = SQL_MODES.index(str(setting.value))
            elif setting.get_name() == "CUSTTONETX":
                if setting.value:
                    _mem.txtone = len(TONES)
            elif setting.get_name() == "CUSTTONERX":
                if setting.value:
                    _mem.rxtone = len(TONES)
            else:
                setattr(_mem, setting.get_name(), setting.value)

        return mem

    def get_settings(self):
        allGroups = RadioSettings()

        _settings = self._memobj.settings

        basic = RadioSettingGroup("basic", "Basic")
        basic.append(
            RadioSetting(
                "welcome",
                "Welcome Message",
                RadioSettingValueString(
                    0,
                    7,
                    _filter(self._memobj.welcome,
                            chirp_common.CHARSET_ASCII))))
        basic.append(
            RadioSetting(
                "display",
                "Display",
                RadioSettingValueList(DISPLAY, DISPLAY[_settings.display])))
        basic.append(
            RadioSetting(
                "disp_chan_lock",
                "Lock Chan# Disp",
                RadioSettingValueBoolean(_settings.disp_chan_lock)))
        basic.append(
            RadioSetting(
                "scan_tot",
                "Scan Pause Time",
                RadioSettingValueList(
                    SCAN_PAUSES,
                    SCAN_PAUSES[_settings.scan_tot])))
        basic.append(
            RadioSetting(
                "scan_mode",
                "Scan Mode",
                RadioSettingValueList(
                    SCAN_MODES,
                    SCAN_MODES[_settings.scan_mode])))
        basic.append(
            RadioSetting(
                "ptt_tot",
                "Talk Timeout",
                RadioSettingValueList(TOT, TOT[_settings.ptt_tot])))
        basic.append(
            RadioSetting(
                "apo",
                "Auto Power Off",
                RadioSettingValueList(APO, APO[_settings.apo])))
        basic.append(
            RadioSetting(
                "beep_on",
                "Beep",
                RadioSettingValueBoolean(_settings.beep_on)))
        basic.append(
            RadioSetting(
                "beep_hi",
                "Beep Volume",
                RadioSettingValueList(BEEP_VOL, BEEP_VOL[_settings.beep_hi])))
        basic.append(
            RadioSetting(
                "ext_spk_on",
                "External Speaker",
                RadioSettingValueBoolean(_settings.ext_spk_on)))
        basic.append(
            RadioSetting(
                "deputy_mute",
                "Deputy Chan Mute",
                RadioSettingValueList(
                    DEPUTY_CHAN_MUTES,
                    DEPUTY_CHAN_MUTES[_settings.deputy_mute])))
        basic.append(
            RadioSetting(
                "tail_elim_type",
                "Tail Eliminator",
                RadioSettingValueList(
                    TAIL_ELIM_TYPES,
                    TAIL_ELIM_TYPES[_settings.tail_elim_type])))
        basic.append(
            RadioSetting(
                "no_tone_elim_tail",
                "Eliminate SQL Tail When\nNo Tone Signaling",
                RadioSettingValueBoolean(_settings.no_tone_elim_tail)))
        allGroups.append(basic)

        sqls = RadioSettingGroup("sql", "Squelch Levels")
        sqls.append(
            RadioSetting(
                "sql_left",
                "Left Squelch",
                RadioSettingValueInteger(0, 20, _settings.sql_left)))
        sqls.append(
            RadioSetting(
                "sql_right",
                "Right Squelch",
                RadioSettingValueInteger(0, 20, _settings.sql_right)))
        sqls.append(
            RadioSetting(
                "rfsql_left",
                "Left RF Squelch",
                RadioSettingValueList(
                    RF_SQUELCHES,
                    RF_SQUELCHES[_settings.rfsql_left])))
        sqls.append(
            RadioSetting(
                "rfsql_right",
                "Right RF Squelch",
                RadioSettingValueList(
                    RF_SQUELCHES,
                    RF_SQUELCHES[_settings.rfsql_right])))
        allGroups.append(sqls)

        keys = RadioSettingGroup("keys", "Keys / Buttons")
        keys.append(
            RadioSetting(
                "sql_key_mode",
                "SQL Key Mode",
                RadioSettingValueList(
                    SQL_BTN_MODES,
                    SQL_BTN_MODES[_settings.sql_key_mode])))
        keys.append(
            RadioSetting(
                "key_a_func",
                "PA Key Function",
                RadioSettingValueList(
                    KEY_FUNCS,
                    KEY_FUNCS[_settings.key_a_func])))
        keys.append(
            RadioSetting(
                "key_b_func",
                "PB Key Function",
                RadioSettingValueList(
                    KEY_FUNCS,
                    KEY_FUNCS[_settings.key_b_func])))
        keys.append(
            RadioSetting(
                "key_c_func",
                "PC Key Function",
                RadioSettingValueList(
                    KEY_FUNCS,
                    KEY_FUNCS[_settings.key_c_func])))
        keys.append(
            RadioSetting(
                "key_d_func",
                "PD Key Function",
                RadioSettingValueList(
                    KEY_FUNCS,
                    KEY_FUNCS[_settings.key_d_func])))
        keys.append(
            RadioSetting(
                "key_bright",
                "Key Light Level",
                _RadioSettingValueOffsetInt(
                    1,
                    32,
                    _settings.key_bright + 1,
                    1,
                    -1)))
        keys.append(
            RadioSetting(
                "ptt_lockout",
                "PTT Key Lock",
                RadioSettingValueList(
                    PTT_KEY_LOCKS,
                    PTT_KEY_LOCKS[_settings.ptt_lockout])))
        keys.append(
            RadioSetting(
                "keypad_lock",
                "Keypad Lockout",
                RadioSettingValueBoolean(_settings.keypad_lock)))
        allGroups.append(keys)

        bgColor = RadioSettingGroup("bgColor", "Background Color")
        bgColor.append(
            RadioSetting(
                "bg_red",
                "Red",
                _RadioSettingValueOffsetInt(
                    1,
                    32,
                    _settings.bg_red + 1,
                    1,
                    -1)))
        bgColor.append(
            RadioSetting(
                "bg_green",
                "Green",
                _RadioSettingValueOffsetInt(
                    1,
                    32,
                    _settings.bg_green + 1,
                    1,
                    -1)))
        bgColor.append(
            RadioSetting(
                "bg_blue",
                "Blue",
                _RadioSettingValueOffsetInt(
                    1,
                    32,
                    _settings.bg_blue + 1,
                    1,
                    -1)))
        allGroups.append(bgColor)

        pwdGroup = RadioSettingGroup("pwdGroup", "Boot Password")
        pwdGroup.append(
            RadioSetting(
                "boot_pw_on",
                "Enable",
                RadioSettingValueBoolean(_settings.boot_pw_on)))
        pwdGroup.append(
            RadioSetting(
                "boot_password",
                "Password (digits)",
                RadioSettingValueString(
                    0,
                    7,
                    _filter(_settings.boot_password, string.digits),
                    False,
                    string.digits)))
        allGroups.append(pwdGroup)

        advanced = RadioSettingGroup("advanced", "Advanced")
        advanced.append(
            RadioSetting(
                "tbst_freq",
                "TBST Freq",
                RadioSettingValueList(
                    TBST_FREQS,
                    TBST_FREQS[_settings.tbst_freq])))
        advanced.append(
            RadioSetting(
                "long_key_time",
                "Long Key Time",
                RadioSettingValueList(
                    LONG_KEY_TIMES,
                    LONG_KEY_TIMES[int((_settings.long_key_time-100) / 50)])))
        advanced.append(
            RadioSetting(
                "clk_shift",
                "CLK Shift",
                RadioSettingValueBoolean(_settings.clk_shift)))
        advanced.append(
            RadioSetting(
                "alt_key_mode",
                "Keypad Alt Mode",
                RadioSettingValueBoolean(_settings.alt_key_mode)))
        advanced.append(
            RadioSetting(
                "no_init",
                "Prohibit Initialization",
                RadioSettingValueBoolean(_settings.no_init)))
        allGroups.append(advanced)

        hyperGroups = [
            RadioSettingGroup("hyper1", "Hyper 1"),
            RadioSettingGroup("hyper2", "Hyper 2")
        ]
        for idx in range(0, len(hyperGroups)):
            _hyperSettings = _settings.hyper_settings[idx]
            hyperGroup = hyperGroups[idx]
            hyperGroup.append(
                RadioSetting(
                    "main",
                    "Main Band",
                    RadioSettingValueList(
                        HYPER_MAINS,
                        HYPER_MAINS[_hyperSettings.main])))
            hyperGroup.append(
                RadioSetting(
                    "sub_display",
                    "Sub Display",
                    RadioSettingValueList(
                        HYPER_SUB_DISPLAYS,
                        HYPER_SUB_DISPLAYS[_hyperSettings.sub_display])))
            hyperGroup.append(
                RadioSetting(
                    "spkr_mode",
                    "Speakers",
                    RadioSettingValueList(
                        HYPER_SPKR_MODES,
                        HYPER_SPKR_MODES[_hyperSettings.spkr_mode])))
            hyperGroup.append(
                RadioSetting(
                    "vfo_band_edge",
                    "VFO Band Lock",
                    RadioSettingValueBoolean(_hyperSettings.vfo_band_edge)))
            hyperGroup.append(
                RadioSetting(
                    "auto_am",
                    "Auto AM",
                    RadioSettingValueBoolean(_hyperSettings.auto_am)))
            hyperGroup.append(
                RadioSetting(
                    "auto_rpt_shift",
                    "Auto Repeater Shift",
                    RadioSettingValueBoolean(_hyperSettings.auto_rpt_shift)))
            hyperGroup.append(
                RadioSetting(
                    "auto_hyper_save",
                    "Auto Hyper Save",
                    RadioSettingValueBoolean(_hyperSettings.auto_hyper_save)))
            hyperGroup.append(
                RadioSetting(
                    "vfo_tracked",
                    "VFO Tracking",
                    RadioSettingValueBoolean(_hyperSettings.vfo_tracked)))
            hyperGroup.append(
                RadioSetting(
                    "left_mode",
                    "Left Mode",
                    RadioSettingValueList(
                        HYPER_MODES,
                        HYPER_MODES[_hyperSettings.left_mode])))
            hyperGroup.append(
                RadioSetting(
                    "right_mode",
                    "Right Mode",
                    RadioSettingValueList(
                        HYPER_MODES,
                        HYPER_MODES[_hyperSettings.right_mode])))
            hyperGroup.append(
                RadioSetting(
                    "left_channel",
                    "Left Channel",
                    _RadioSettingValueOffsetInt(
                        1,
                        750,
                        _hyperSettings.left_channel + 1,
                        1,
                        -1)))
            hyperGroup.append(
                RadioSetting(
                    "right_channel",
                    "Right Channel",
                    _RadioSettingValueOffsetInt(
                        1,
                        750,
                        _hyperSettings.right_channel + 1,
                        1,
                        -1)))
            hyperGroup.append(
                RadioSetting(
                    "left_vfo_band",
                    "Left VFO Band",
                    RadioSettingValueList(
                        HYPER_L_VFOS,
                        HYPER_L_VFOS[_hyperSettings.left_vfo_band])))
            hyperGroup.append(
                RadioSetting(
                    "right_vfo_band",
                    "Right VFO Band",
                    RadioSettingValueList(
                        HYPER_R_VFOS,
                        HYPER_R_VFOS[_hyperSettings.right_vfo_band])))
            hyperGroup.append(
                RadioSetting(
                    "left_work_bank",
                    "Left Work Bank",
                    RadioSettingValueList(
                        BANK_CHOICES,
                        BANK_CHOICES[_hyperSettings.left_work_bank])))
            hyperGroup.append(
                RadioSetting(
                    "right_work_bank",
                    "Right Work Bank",
                    RadioSettingValueList(
                        BANK_CHOICES,
                        BANK_CHOICES[_hyperSettings.right_work_bank])))
            hyperGroup.append(
                RadioSetting(
                    "left_bank_mode",
                    "Left Bank Mode",
                    RadioSettingValueList(
                        HYPER_BANK_MODES,
                        HYPER_BANK_MODES[_hyperSettings.left_bank_mode])))
            hyperGroup.append(
                RadioSetting(
                    "right_bank_mode",
                    "Right Bank Mode",
                    RadioSettingValueList(
                        HYPER_BANK_MODES,
                        HYPER_BANK_MODES[_hyperSettings.right_bank_mode])))
            hyperGroup.append(
                RadioSetting(
                    "left_bank_sw",
                    "Left Bank Switch",
                    RadioSettingValueBoolean(_hyperSettings.left_bank_sw)))
            hyperGroup.append(
                RadioSetting(
                    "right_bank_sw",
                    "Right Bank Switch",
                    RadioSettingValueBoolean(_hyperSettings.right_bank_sw)))
            for bank in BANK_CHOICES[0:-1]:
                hyperGroup.append(
                    RadioSetting(
                        "link_bank_%s" % bank,
                        "Bank %s" % bank,
                        RadioSettingValueBoolean(
                            getattr(_hyperSettings, "link_bank_%s" % bank))))
            allGroups.append(hyperGroup)

        notes = RadioSettingGroup("notes", "Comm Notes")
        for idx in range(0, 100):
            notes.append(
                RadioSetting(
                    "note_call_ids.%d" % idx,
                    "Call ID %d (5 digits)" % idx,
                    RadioSettingValueString(
                        0,
                        5,
                        _filter(self._memobj.note_call_ids[idx].call_id,
                                string.digits)[idx*8:idx*8+5],
                        False,
                        string.digits)))
            notes.append(
                RadioSetting(
                    "note_names.%d" % idx,
                    "Name %d (7 ALPHAnumeric)" % idx,
                    RadioSettingValueString(
                        0,
                        7,
                        _filter(self._memobj.note_names[idx].name,
                                chirp_common.CHARSET_UPPER_NUMERIC)
                        [idx*8:idx*8+7],
                        True,
                        chirp_common.CHARSET_UPPER_NUMERIC)))
        allGroups.append(notes)

        _emergency = self._memobj.emergency
        emer = RadioSettingGroup("emergency", "Emergency Info")
        emer.append(
            RadioSetting(
                "mode",
                "Alarm Mode",
                RadioSettingValueList(
                    EMER_MODES,
                    EMER_MODES[_emergency.mode])))
        emer.append(
            RadioSetting(
                "eni_type",
                "ENI Type",
                RadioSettingValueList(
                    EMER_ENI_TYPES,
                    EMER_ENI_TYPES[_emergency.eni_type])))
        emer.append(
            RadioSetting(
                "emergency.id",
                "DTMF ID",
                RadioSettingValueList(
                    DTMF_SLOTS,
                    DTMF_SLOTS[_emergency.id]
                    if EMER_ENI_TYPES[_emergency.eni_type] == 'DTMF'
                    else DTMF_SLOTS[0])))
        emer.append(
            RadioSetting(
                "5ToneId",
                "5Tone ID",
                RadioSettingValueInteger(
                    0,
                    99,
                    _emergency.id
                    if EMER_ENI_TYPES[_emergency.eni_type] == '5Tone'
                    else 0)))
        emer.append(
            RadioSetting(
                "alarm_time",
                "Alarm Time (sec)",
                RadioSettingValueInteger(1, 255, _emergency.alarm_time)))
        emer.append(
            RadioSetting(
                "tx_time",
                "TX Duration (sec)",
                RadioSettingValueInteger(1 if _emergency.tx_time > 0 else 0,
                                         255,
                                         _emergency.tx_time)))
        emer.append(
            RadioSetting(
                "rx_time",
                "RX Duration (sec)",
                RadioSettingValueInteger(1 if _emergency.rx_time > 0 else 0,
                                         255,
                                         _emergency.rx_time)))
        emer.append(
            RadioSetting(
                "chan_select",
                "ENI Channel",
                RadioSettingValueList(
                    EMER_CHAN_SEL,
                    EMER_CHAN_SEL[_emergency.chan_select])))
        emer.append(
            RadioSetting(
                "channel",
                "Assigned Channel",
                _RadioSettingValueOffsetInt(
                    1,
                    750,
                    _emergency.channel + 1,
                    1,
                    -1)))
        emer.append(
            RadioSetting(
                "cycle",
                "Cycle",
                RadioSettingValueList(
                    EMER_CYCLES,
                    EMER_CYCLES[_emergency.cycle])))
        allGroups.append(emer)

        dtmfGroup = RadioSettingGroup("dtmf", "DTMF")
        for idx in range(0, 16):
            dtmfGroup.append(
                RadioSetting(
                    "dtmf_encodings.%d" % idx,
                    "Encode Tone %s" % DTMF_SLOTS[idx],
                    RadioSettingValueString(
                        0,
                        12,
                        _dtmf_decode(
                            self._memobj.dtmf_encodings[idx].dtmf,
                            self._memobj.dtmf_encodings[idx].num_tones),
                        False,
                        DTMF_CHARS)))
        dtmfGroup.append(
            RadioSetting(
                "dtmf_preload_time",
                "Encode Preload Time",
                RadioSettingValueList(
                    DTMF_PRELOADS,
                    DTMF_PRELOADS[_settings.dtmf_preload_time])))
        dtmfGroup.append(
            RadioSetting(
                "dtmf_speed",
                "Speed",
                RadioSettingValueList(
                    DTMF_SPEEDS,
                    DTMF_SPEEDS[_settings.dtmf_speed])))
        dtmfGroup.append(
            RadioSetting(
                "dtmf_interval_char",
                "Interval Character",
                RadioSettingValueList(
                    list(DTMF_INTCHARS),
                    DTMF_CHARS[_settings.dtmf_interval_char])))
        dtmfGroup.append(
            RadioSetting(
                "dtmf_group_code",
                "Group Code",
                RadioSettingValueList(
                    DTMF_GROUPS,
                    DTMF_GROUPS[_settings.dtmf_group_code-0x0a]
                    if _settings.dtmf_group_code < 0xff
                    else DTMF_GROUPS[6])))
        dtmfGroup.append(
            RadioSetting(
                "dtmf_dec_resp",
                "Decode Response",
                RadioSettingValueList(
                    DTMF_RESPONSES,
                    DTMF_RESPONSES[_settings.dtmf_dec_resp])))
        dtmfGroup.append(
            RadioSetting(
                "dtmf_first_dig_time",
                "1st Digit Time (ms)",
                RadioSettingValueInteger(
                    0,
                    2500,
                    _settings.dtmf_first_dig_time * 10,
                    10)))
        dtmfGroup.append(
            RadioSetting(
                "dtmf_auto_rst_time",
                "Auto Reset Time (0-25 sec by 0.1)",
                RadioSettingValueFloat(
                    0,
                    25,
                    _settings.dtmf_auto_rst_time / 10,
                    0.1,
                    1)))
        dtmfGroup.append(
            RadioSetting(
                "dtmf_self_id",
                "Self ID",
                RadioSettingValueString(
                    3,
                    3,
                    (DTMF_CHARS[_settings.dtmf_self_id[0] & 0x0F] +
                     DTMF_CHARS[_settings.dtmf_self_id[1] & 0x0F] +
                     DTMF_CHARS[_settings.dtmf_self_id[2] & 0x0F]),
                    True,
                    DTMF_CHARS)))
        dtmfGroup.append(
            RadioSetting(
                "dtmf_sidetone_on",
                "Use Side Tone",
                RadioSettingValueBoolean(_settings.dtmf_sidetone_on)))
        dtmfGroup.append(
            RadioSetting(
                "dtmf_enc_delay",
                "Encode Delay (ms)",
                RadioSettingValueInteger(
                    10,
                    2500,
                    _settings.dtmf_enc_delay * 10,
                    10)))
        dtmfGroup.append(
            RadioSetting(
                "dtmf_pttid_delay",
                "PTT ID Delay",
                RadioSettingValueList(
                    DTMF_PTTDELAYS,
                    DTMF_PTTDELAYS[_settings.dtmf_pttid_delay-4]
                    if _settings.dtmf_pttid_delay >= 5
                    else DTMF_PTTDELAYS[0])))
        dtmfGroup.append(
            RadioSetting(
                "dtmf_kill",
                "Remote Kill",
                RadioSettingValueString(
                    0,
                    7,
                    _dtmf_decode(_settings.dtmf_kill.dtmf,
                                 _settings.dtmf_kill.num_tones),
                    False,
                    DTMF_CHARS)))
        dtmfGroup.append(
            RadioSetting(
                "dtmf_stun",
                "Remote Stun",
                RadioSettingValueString(
                    0,
                    7,
                    _dtmf_decode(_settings.dtmf_stun.dtmf,
                                 _settings.dtmf_stun.num_tones),
                    False,
                    DTMF_CHARS)))
        dtmfGroup.append(
            RadioSetting(
                "dtmf_pttid_bot",
                "PTT ID Start (BOT)",
                RadioSettingValueString(
                    0,
                    12,
                    _dtmf_decode(self._memobj.dtmf_pttid_bot.dtmf,
                                 self._memobj.dtmf_pttid_bot.num_tones),
                    False,
                    DTMF_CHARS)))
        dtmfGroup.append(
            RadioSetting(
                "dtmf_pttid_eot",
                "PTT ID Start (EOT)",
                RadioSettingValueString(
                    0,
                    12,
                    _dtmf_decode(self._memobj.dtmf_pttid_eot.dtmf,
                                 self._memobj.dtmf_pttid_eot.num_tones),
                    False,
                    DTMF_CHARS)))
        allGroups.append(dtmfGroup)

        # TODO: 2-Tone settings
        # TODO: 5-Tone settings

        return allGroups

    def _set_hyper_settings(self, settings, _hyperSettings):
        for element in settings:
            name = element.get_name()
            if not isinstance(element, RadioSetting):
                self._set_hyper_settings(element, _hyperSettings)
                continue
            elif isinstance(element.value, _RadioSettingValueOffsetInt):
                setattr(_hyperSettings,
                        name,
                        int(element.value) + element.value.offset)
            else:
                setattr(_hyperSettings, name, element.value)

    def _set_pwd_settings(self, settings):
        _settings = self._memobj.settings
        use_pw = pwd = None
        for element in settings:
            name = element.get_name()
            if not isinstance(element, RadioSetting):
                self._set_pwd_settings(element)
                continue
            elif name == "boot_pw_on":
                use_pw = bool(element.value)
            elif name == "boot_password":
                pwd = _filter(str(element.value), string.digits)
        if (pwd is not None and (len(pwd) or use_pw is False)):
            setattr(_settings, "boot_pw_on", use_pw and 1 or 0)
            pwd = pwd + "\x00"*(7-len(pwd))
            setattr(_settings, "boot_password", pwd)
        elif (use_pw is not None and use_pw is True):
            raise errors.InvalidValueError("Password may not be enabled "
                                           "without being set.")

    def _set_notes_settings(self, settings):
        _noteCalls = self._memobj.note_call_ids
        _noteNames = self._memobj.note_names
        for element in settings:
            name = str(element.get_name())
            if not isinstance(element, RadioSetting):
                self._set_notes_settings(element)
                continue
            elif name.startswith("note_call_ids."):
                idx = int(name[len("note_call_ids."):])
                setattr(self._memobj.note_call_ids[idx],
                        "call_id",
                        str(element.value).ljust(5, "\x00"))
            elif name.startswith("note_names."):
                idx = int(name[len("note_names."):])
                setattr(self._memobj.note_names[idx],
                        "name",
                        str(element.value).ljust(7))

    def _set_emer_settings(self, settings):
        _emergency = self._memobj.emergency
        dtmf_id = fivetone_id = None
        for element in settings:
            name = str(element.get_name())
            if not isinstance(element, RadioSetting):
                self._set_emer_settings(element)
                continue
            elif name == "emergency.id":
                dtmf_id = element.value
            elif name == "5ToneId":
                fivetone_id = element.value
            elif name == "channel":
                zb_mem_num = int(element.value) + element.value.offset
                _mem, _flg = self._get_memobjs(zb_mem_num + 1)
                if _flg.unused:
                    raise errors.InvalidValueError("Assigned emergency ENI "
                                                   "channel is empty.")
                else:
                    setattr(_emergency, name, zb_mem_num)
            elif isinstance(element.value, _RadioSettingValueOffsetInt):
                setattr(_emergency,
                        name,
                        int(element.value) + element.value.offset)
            else:
                setattr(_emergency, name, element.value)
        if EMER_ENI_TYPES[_emergency.eni_type] == 'DTMF':
            setattr(_emergency, "id", dtmf_id)
        elif EMER_ENI_TYPES[_emergency.eni_type] == '5Tone':
            setattr(_emergency, "id", fivetone_id)

    def _set_dtmf_settings(self, settings):
        def _setDtmf(dtmfStruct, newVal):
            dtmfStruct.num_tones.set_value(len(newVal))
            dtmfStruct.dtmf.set_value(_dtmf_encode(newVal,
                                                   len(dtmfStruct.dtmf)))

        for element in settings:
            name = str(element.get_name())
            if not isinstance(element, RadioSetting):
                self._set_dtmf_settings(element)
                continue
            if name.startswith("dtmf_encodings."):
                idx = int(name[len("dtmf_encodings."):])
                _setDtmf(self._memobj.dtmf_encodings[idx], str(element.value))
            elif name == "dtmf_interval_char":
                setattr(self._memobj.settings.dtmf_interval_char,
                        name,
                        DTMF_CHARS.index(str(element.value)))
            elif name == "dtmf_group_code":
                setattr(self._memobj.settings,
                        name,
                        int(element.value)+0x0a
                        if int(element.value) != 6
                        else 0xff)
            elif name == "dtmf_first_dig_time" or name == "dtmf_enc_delay":
                setattr(self._memobj.settings, name, int(element.value) / 10)
            elif name == "dtmf_auto_rst_time":
                setattr(self._memobj.settings, name, int(element.value) * 10)
            elif name == "dtmf_self_id":
                newStr = str(element.value)
                newVal = []
                for charIdx in range(0, 3):
                    newVal.append(DTMF_CHARS.index(newStr[charIdx]))
                setattr(self._memobj.settings, name, newVal)
            elif name == "dtmf_pttid_delay":
                setattr(self._memobj.settings,
                        name,
                        DTMF_PTTDELAYS.index(str(element.value))+4
                        if element.value != DTMF_PTTDELAYS[0]
                        else 0)
            elif name == "dtmf_kill" or name == "dtmf_stun":
                _setDtmf(getattr(self._memobj.settings, name),
                         str(element.value))
            elif name == "dtmf_pttid_bot" or name == "dtmf_pttid_eot":
                _setDtmf(getattr(self._memobj, name), str(element.value))
            else:
                setattr(self._memobj.settings, name, element.value)

    def set_settings(self, settings):
        _root = self._memobj
        _settings = self._memobj.settings
        for element in settings:
            name = element.get_name()
            if isinstance(element, RadioSettingGroup):
                if name == "hyper1":
                    self._set_hyper_settings(
                        element,
                        _settings.hyper_settings[0])
                    continue
                elif name == "hyper2":
                    self._set_hyper_settings(
                        element,
                        _settings.hyper_settings[1])
                    continue
                elif name == "pwdGroup":
                    self._set_pwd_settings(element)
                    continue
                elif name == "notes":
                    self._set_notes_settings(element)
                    continue
                elif name == "emergency":
                    self._set_emer_settings(element)
                    continue
                elif name == "dtmf":
                    self._set_dtmf_settings(element)
                    continue
            if not isinstance(element, RadioSetting):
                self.set_settings(element)
                continue
            if name == "ignore":
                LOG.debug("*** ignore: %s" % str(element.value))
            elif name == "welcome":
                setattr(_root, "welcome", element.value)
            elif name == "long_key_time":
                setattr(_settings, name, 100 + 50 * int(element.value))
            elif isinstance(element.value, _RadioSettingValueOffsetInt):
                setattr(_settings, name, (int(element.value) +
                                          element.value.offset))
            else:
                setattr(_settings, name, element.value)

    @classmethod
    def match_model(cls, filedata, filename):
        return cls._file_ident in filedata[0x40:0x4D]



# Copyright 2008 Dan Smith <dsmith@danplanet.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from builtins import bytes

import base64
import json
import inspect
import logging
import math
import re
import sys

from chirp import errors, memmap, CHIRP_VERSION

LOG = logging.getLogger(__name__)

SEPCHAR = ","

# 50 Tones
TONES = (
    67.0, 69.3, 71.9, 74.4, 77.0, 79.7, 82.5,
    85.4, 88.5, 91.5, 94.8, 97.4, 100.0, 103.5,
    107.2, 110.9, 114.8, 118.8, 123.0, 127.3,
    131.8, 136.5, 141.3, 146.2, 151.4, 156.7,
    159.8, 162.2, 165.5, 167.9, 171.3, 173.8,
    177.3, 179.9, 183.5, 186.2, 189.9, 192.8,
    196.6, 199.5, 203.5, 206.5, 210.7, 218.1,
    225.7, 229.1, 233.6, 241.8, 250.3, 254.1,
)

OLD_TONES = tuple(set(TONES) - set([159.8, 165.5, 171.3, 177.3, 183.5, 189.9,
                                    196.6, 199.5, 206.5, 229.1, 254.1]))


def VALIDTONE(v):
    return isinstance(v, float) and 50 < v < 300


# 104 DTCS Codes
DTCS_CODES = (
    23,  25,  26,  31,  32,  36,  43,  47,  51,  53,  54,
    65,  71,  72,  73,  74,  114, 115, 116, 122, 125, 131,
    132, 134, 143, 145, 152, 155, 156, 162, 165, 172, 174,
    205, 212, 223, 225, 226, 243, 244, 245, 246, 251, 252,
    255, 261, 263, 265, 266, 271, 274, 306, 311, 315, 325,
    331, 332, 343, 346, 351, 356, 364, 365, 371, 411, 412,
    413, 423, 431, 432, 445, 446, 452, 454, 455, 462, 464,
    465, 466, 503, 506, 516, 523, 526, 532, 546, 565, 606,
    612, 624, 627, 631, 632, 654, 662, 664, 703, 712, 723,
    731, 732, 734, 743, 754,
)

# 512 Possible DTCS Codes
ALL_DTCS_CODES = tuple([((a * 100) + (b * 10) + c)
                        for a in range(0, 8)
                        for b in range(0, 8)
                        for c in range(0, 8)])

CROSS_MODES = (
    "Tone->Tone",
    "DTCS->",
    "->DTCS",
    "Tone->DTCS",
    "DTCS->Tone",
    "->Tone",
    "DTCS->DTCS",
    "Tone->"
)

# This is the "master" list of modes, and in general things should not be
# added here without significant consideration. These must remain stable and
# universal to allow importing memories between different radio vendors and
# models.
MODES = ("WFM", "FM", "NFM", "AM", "NAM", "DV", "USB", "LSB", "CW", "RTTY",
         "DIG", "PKT", "NCW", "NCWR", "CWR", "P25", "Auto", "RTTYR",
         "FSK", "FSKR", "DMR", "DN")

TONE_MODES = (
    "",
    "Tone",
    "TSQL",
    "DTCS",
    "DTCS-R",
    "TSQL-R",
    "Cross",
)

TUNING_STEPS = (
    5.0, 6.25, 10.0, 12.5, 15.0, 20.0, 25.0, 30.0, 50.0, 100.0,
    125.0, 200.0,
    # Need to fix drivers using this list as an index!
    9.0, 1.0, 2.5,
)

# These are the default for RadioFeatures.valid_tuning_steps
COMMON_TUNING_STEPS = (5.0, 10.0, 15.0, 20.0, 25.0, 30.0, 50.0, 100.0)

SKIP_VALUES = ("", "S", "P")

CHARSET_UPPER_NUMERIC = "ABCDEFGHIJKLMNOPQRSTUVWXYZ 1234567890"
CHARSET_ALPHANUMERIC = \
    "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz 1234567890"
CHARSET_ASCII = "".join([chr(x) for x in range(ord(" "), ord("~") + 1)])
CHARSET_1252 = bytes(
    [x for x in range(0x20, 0x100)
     if x not in [0x7F, 0x81, 0x8D, 0x8F, 0x90, 0x9D, 0xA0, 0xAD]]
).decode('cp1252')

# http://aprs.org/aprs11/SSIDs.txt
APRS_SSID = (
    "0 Your primary station usually fixed and message capable",
    "1 generic additional station, digi, mobile, wx, etc",
    "2 generic additional station, digi, mobile, wx, etc",
    "3 generic additional station, digi, mobile, wx, etc",
    "4 generic additional station, digi, mobile, wx, etc",
    "5 Other networks (Dstar, Iphones, Androids, Blackberry's etc)",
    "6 Special activity, Satellite ops, camping or 6 meters, etc",
    "7 walkie talkies, HT's or other human portable",
    "8 boats, sailboats, RV's or second main mobile",
    "9 Primary Mobile (usually message capable)",
    "10 internet, Igates, echolink, winlink, AVRS, APRN, etc",
    "11 balloons, aircraft, spacecraft, etc",
    "12 APRStt, DTMF, RFID, devices, one-way trackers*, etc",
    "13 Weather stations",
    "14 Truckers or generally full time drivers",
    "15 generic additional station, digi, mobile, wx, etc")
APRS_POSITION_COMMENT = (
    "off duty", "en route", "in service", "returning", "committed",
    "special", "priority", "custom 0", "custom 1", "custom 2", "custom 3",
    "custom 4", "custom 5", "custom 6", "EMERGENCY")
# http://aprs.org/symbols/symbolsX.txt
APRS_SYMBOLS = (
    "Police/Sheriff", "[reserved]", "Digi", "Phone", "DX Cluster",
    "HF Gateway", "Small Aircraft", "Mobile Satellite Groundstation",
    "Wheelchair", "Snowmobile", "Red Cross", "Boy Scouts", "House QTH (VHF)",
    "X", "Red Dot", "0 in Circle", "1 in Circle", "2 in Circle",
    "3 in Circle", "4 in Circle", "5 in Circle", "6 in Circle", "7 in Circle",
    "8 in Circle", "9 in Circle", "Fire", "Campground", "Motorcycle",
    "Railroad Engine", "Car", "File Server", "Hurricane Future Prediction",
    "Aid Station", "BBS or PBBS", "Canoe", "[reserved]", "Eyeball",
    "Tractor/Farm Vehicle", "Grid Square", "Hotel", "TCP/IP", "[reserved]",
    "School", "PC User", "MacAPRS", "NTS Station", "Balloon", "Police", "TBD",
    "Recreational Vehicle", "Space Shuttle", "SSTV", "Bus", "ATV",
    "National WX Service Site", "Helicopter", "Yacht/Sail Boat", "WinAPRS",
    "Human/Person", "Triangle", "Mail/Postoffice", "Large Aircraft",
    "WX Station", "Dish Antenna", "Ambulance", "Bicycle",
    "Incident Command Post", "Dual Garage/Fire Dept", "Horse/Equestrian",
    "Fire Truck", "Glider", "Hospital", "IOTA", "Jeep", "Truck", "Laptop",
    "Mic-Repeater", "Node", "Emergency Operations Center", "Rover (dog)",
    "Grid Square above 128m", "Repeater", "Ship/Power Boat", "Truck Stop",
    "Truck (18 wheeler)", "Van", "Water Station", "X-APRS", "Yagi at QTH",
    "TDB", "[reserved]"
)


def watts_to_dBm(watts):
    """Converts @watts in watts to dBm"""
    return 10 * math.log10(watts) + 30


def dBm_to_watts(dBm):
    """Converts @dBm from dBm to watts"""
    return round(math.pow(10, dBm / 10) / 1000, 1)


class PowerLevel:
    """Represents a power level supported by a radio"""

    def __init__(self, label, watts=0, dBm=0):
        if watts:
            dBm = watts_to_dBm(watts)
        self._power = float(dBm)
        self._label = label

    def __str__(self):
        return str(self._label)

    def __int__(self):
        return int(self._power)

    def __float__(self):
        return self._power

    def __sub__(self, val):
        return float(self) - float(val)

    def __add__(self, val):
        return float(self) + float(val)

    def __eq__(self, val):
        if val is not None:
            return float(self) == float(val)
        return False

    def __lt__(self, val):
        return float(self) < float(val)

    def __gt__(self, val):
        return float(self) > float(val)

    def __bool__(self):
        return int(self) != 0

    def __repr__(self):
        return "%s (%i dBm)" % (self._label, self._power)


class AutoNamedPowerLevel(PowerLevel):
    """A power level that is simply named by its value in watts"""

    def __init__(self, watts):
        fmt = ('%iW' if watts >= 10 else '%.1fW')
        super().__init__(fmt % watts, watts=watts)


def parse_power(powerstr):
    if powerstr.isdigit():
        # All digits means watts
        watts = int(powerstr)
    else:
        match = re.match(r'^\s*([0-9.]+)\s*([Ww]?)\s*$', powerstr)
        if not match:
            raise ValueError('Invalid power specification: %r' % powerstr)
        if match.group(2).lower() in ('', 'w'):
            watts = float(match.group(1))
        else:
            raise ValueError('Unknown power units in %r' % powerstr)

    return AutoNamedPowerLevel(watts)


def parse_freq(freqstr):
    """Parse a frequency string and return the value in integral Hz"""
    freqstr = freqstr.strip()
    if freqstr == "":
        return 0
    elif freqstr.endswith(" MHz"):
        return parse_freq(freqstr.split(" ")[0])
    elif freqstr.endswith(" kHz"):
        return int(freqstr.split(" ")[0]) * 1000

    if "." in freqstr:
        mhz, khz = freqstr.split(".")
        if mhz == "":
            mhz = 0
        khz = khz.ljust(6, "0")
        if len(khz) > 6:
            raise ValueError("Invalid kHz value: %s", khz)
        mhz = int(mhz) * 1000000
        khz = int(khz)
    else:
        mhz = int(freqstr) * 1000000
        khz = 0

    return mhz + khz


def format_freq(freq):
    """Format a frequency given in Hz as a string"""

    return "%i.%06i" % (freq / 1000000, freq % 1000000)


class ImmutableValueError(ValueError):
    pass


class Memory:
    """Base class for a single radio memory"""
    freq = 0
    number = 0
    extd_number = ""
    name = ""
    vfo = 0
    rtone = 88.5
    ctone = 88.5
    dtcs = 23
    rx_dtcs = 23
    tmode = ""
    cross_mode = "Tone->Tone"
    dtcs_polarity = "NN"
    skip = ""
    power = None
    duplex = ""
    offset = 600000
    mode = "FM"
    tuning_step = 5.0

    comment = ""

    empty = False

    immutable = []

    # A RadioSettingGroup of additional settings supported by the radio,
    # or an empty list if none
    extra = []

    def __init__(self, number=0, empty=False, name=""):
        self.freq = 0
        self.number = number
        self.extd_number = ""
        self.name = name
        self.vfo = 0
        self.rtone = 88.5
        self.ctone = 88.5
        self.dtcs = 23
        self.rx_dtcs = 23
        self.tmode = ""
        self.cross_mode = "Tone->Tone"
        self.dtcs_polarity = "NN"
        self.skip = ""
        self.power = None
        self.duplex = ""
        self.offset = 600000
        self.mode = "FM"
        self.tuning_step = 5.0

        self.comment = ""

        self.empty = empty

        self.immutable = []

    _valid_map = {
        "rtone":          VALIDTONE,
        "ctone":          VALIDTONE,
        "dtcs":           ALL_DTCS_CODES,
        "rx_dtcs":        ALL_DTCS_CODES,
        "tmode":          TONE_MODES,
        "dtcs_polarity":  ["NN", "NR", "RN", "RR"],
        "cross_mode":     CROSS_MODES,
        "mode":           MODES,
        "duplex":         ["", "+", "-", "split", "off"],
        "skip":           SKIP_VALUES,
        "empty":          [True, False],
        "dv_code":        [x for x in range(0, 100)],
    }

    def __repr__(self):
        ident, vals = self.debug_dump()
        return '<Memory %s: %s>' % (
            ident, ','.join('%s=%r' % item for item in vals))

    def debug_diff(self, other, delim='/'):
        my_ident, my_vals = self.debug_dump()
        my_vals = dict(my_vals)

        om_ident, om_vals = other.debug_dump()
        om_vals = dict(om_vals)

        diffs = []
        if my_ident != om_ident:
            diffs.append('ident=%s%s%s' % (my_ident, delim, om_ident))
        for k in sorted(my_vals.keys() | om_vals.keys()):
            myval = my_vals.get(k, '<missing>')
            omval = om_vals.get(k, '<missing>')
            if myval != omval:
                diffs.append('%s=%r%s%r' % (k, myval, delim, omval))
        return ','.join(diffs)

    def debug_dump(self):
        vals = [(k, v) for k, v in self.__dict__.items()
                if k not in ('extra', 'number', 'extd_number')]
        for extra in self.extra:
            vals.append(('extra.%s' % extra.get_name(), str(extra.value)))
        if self.extd_number:
            ident = '%s(%i)' % (self.extd_number, self.number)
        else:
            ident = str(self.number)
        return ident, vals

    def dupe(self):
        """Return a deep copy of @self"""
        mem = self.__class__()
        for k, v in list(self.__dict__.items()):
            mem.__dict__[k] = v

        return mem

    def clone(self, source):
        """Absorb all of the properties of @source"""
        for k, v in list(source.__dict__.items()):
            self.__dict__[k] = v

    CSV_FORMAT = ["Location", "Name", "Frequency",
                  "Duplex", "Offset", "Tone",
                  "rToneFreq", "cToneFreq", "DtcsCode",
                  "DtcsPolarity", "RxDtcsCode",
                  "CrossMode",
                  "Mode", "TStep",
                  "Skip", "Power", "Comment",
                  "URCALL", "RPT1CALL", "RPT2CALL", "DVCODE"]

    def __setattr__(self, name, val):
        if not hasattr(self, name):
            raise ValueError("No such attribute `%s'" % name)

        if name in self.immutable:
            raise ImmutableValueError("Field %s is not " % name +
                                      "mutable on this memory")

        if name in self._valid_map:
            valid = self._valid_map[name]
            if callable(valid):
                if not valid(val):
                    raise ValueError("`%s' is not a valid value for `%s'" % (
                                         val, name))
            elif val not in self._valid_map[name]:
                raise ValueError("`%s' is not in valid list: %s" %
                                 (val, self._valid_map[name]))

        self.__dict__[name] = val

    def format_freq(self):
        """Return a properly-formatted string of this memory's frequency"""
        return format_freq(self.freq)

    def parse_freq(self, freqstr):
        """Set the frequency from a string"""
        self.freq = parse_freq(freqstr)
        return self.freq

    def __str__(self):
        if self.tmode == "Tone":
            tenc = "*"
        else:
            tenc = " "

        if self.tmode == "TSQL":
            tsql = "*"
        else:
            tsql = " "

        if self.tmode == "DTCS":
            dtcs = "*"
        else:
            dtcs = " "

        if self.duplex == "":
            dup = "/"
        else:
            dup = self.duplex

        return \
            "Memory %s: %s%s%s %s (%s) r%.1f%s c%.1f%s d%03i%s%s [%.2f]" % \
            (self.number if self.extd_number == "" else self.extd_number,
             format_freq(self.freq),
             dup,
             format_freq(self.offset),
             self.mode,
             self.name,
             self.rtone,
             tenc,
             self.ctone,
             tsql,
             self.dtcs,
             dtcs,
             self.dtcs_polarity,
             self.tuning_step)

    def to_csv(self):
        """Return a CSV representation of this memory"""
        return [
            "%i" % self.number,
            "%s" % self.name,
            format_freq(self.freq),
            "%s" % self.duplex,
            format_freq(self.offset),
            "%s" % self.tmode,
            "%.1f" % self.rtone,
            "%.1f" % self.ctone,
            "%03i" % self.dtcs,
            "%s" % self.dtcs_polarity,
            "%03i" % self.rx_dtcs,
            "%s" % self.cross_mode,
            "%s" % self.mode,
            "%.2f" % self.tuning_step,
            "%s" % self.skip,
            "%s" % self.power,
            "%s" % self.comment,
            "", "", "", ""]

    @classmethod
    def _from_csv(cls, _line):
        line = _line.strip()
        if line.startswith("Location"):
            raise errors.InvalidMemoryLocation("Non-CSV line")

        vals = line.split(SEPCHAR)
        if len(vals) < 11:
            raise errors.InvalidDataError("CSV format error " +
                                          "(14 columns expected)")

        if vals[10] == "DV":
            mem = DVMemory()
        else:
            mem = Memory()

        mem.really_from_csv(vals)
        return mem

    def really_from_csv(self, vals):
        """Careful parsing of split-out @vals"""
        try:
            self.number = int(vals[0])
        except Exception:
            raise errors.InvalidDataError(
                "Location '%s' is not a valid integer" % vals[0])

        self.name = vals[1]

        try:
            self.freq = float(vals[2])
        except Exception:
            raise errors.InvalidDataError("Frequency is not a valid number")

        if vals[3].strip() in ["+", "-", ""]:
            self.duplex = vals[3].strip()
        else:
            raise errors.InvalidDataError("Duplex is not +,-, or empty")

        try:
            self.offset = float(vals[4])
        except Exception:
            raise errors.InvalidDataError("Offset is not a valid number")

        self.tmode = vals[5]
        if self.tmode not in TONE_MODES:
            raise errors.InvalidDataError("Invalid tone mode `%s'" %
                                          self.tmode)

        try:
            self.rtone = float(vals[6])
        except Exception:
            raise errors.InvalidDataError("rTone is not a valid number")
        if self.rtone not in TONES:
            raise errors.InvalidDataError("rTone is not valid")

        try:
            self.ctone = float(vals[7])
        except Exception:
            raise errors.InvalidDataError("cTone is not a valid number")
        if self.ctone not in TONES:
            raise errors.InvalidDataError("cTone is not valid")

        try:
            self.dtcs = int(vals[8], 10)
        except Exception:
            raise errors.InvalidDataError("DTCS code is not a valid number")
        if self.dtcs not in DTCS_CODES:
            raise errors.InvalidDataError("DTCS code is not valid")

        try:
            self.rx_dtcs = int(vals[8], 10)
        except Exception:
            raise errors.InvalidDataError("DTCS Rx code is not a valid number")
        if self.rx_dtcs not in DTCS_CODES:
            raise errors.InvalidDataError("DTCS Rx code is not valid")

        if vals[9] in ["NN", "NR", "RN", "RR"]:
            self.dtcs_polarity = vals[9]
        else:
            raise errors.InvalidDataError("DtcsPolarity is not valid")

        if vals[10] in MODES:
            self.mode = vals[10]
        else:
            raise errors.InvalidDataError("Mode is not valid")

        try:
            self.tuning_step = float(vals[11])
        except Exception:
            raise errors.InvalidDataError("Tuning step is invalid")

        try:
            self.skip = vals[12]
        except Exception:
            raise errors.InvalidDataError("Skip value is not valid")

        return True


class DVMemory(Memory):
    """A Memory with D-STAR attributes"""
    dv_urcall = "CQCQCQ"
    dv_rpt1call = ""
    dv_rpt2call = ""
    dv_code = 0

    def __str__(self):
        string = Memory.__str__(self)

        string += " <%s,%s,%s>" % (self.dv_urcall,
                                   self.dv_rpt1call,
                                   self.dv_rpt2call)

        return string

    def to_csv(self):
        return [
            "%i" % self.number,
            "%s" % self.name,
            format_freq(self.freq),
            "%s" % self.duplex,
            format_freq(self.offset),
            "%s" % self.tmode,
            "%.1f" % self.rtone,
            "%.1f" % self.ctone,
            "%03i" % self.dtcs,
            "%s" % self.dtcs_polarity,
            "%s" % self.mode,
            "%.2f" % self.tuning_step,
            "%s" % self.skip,
            "%s" % self.comment,
            "%s" % self.dv_urcall,
            "%s" % self.dv_rpt1call,
            "%s" % self.dv_rpt2call,
            "%i" % self.dv_code]

    def really_from_csv(self, vals):
        Memory.really_from_csv(self, vals)

        self.dv_urcall = vals[15].rstrip()[:8]
        self.dv_rpt1call = vals[16].rstrip()[:8]
        self.dv_rpt2call = vals[17].rstrip()[:8]
        try:
            self.dv_code = int(vals[18].strip())
        except Exception:
            self.dv_code = 0


def FrozenMemory(source):
    class _FrozenMemory(source.__class__):
        def __init__(self, source):
            self.__dict__['_frozen'] = False
            for k, v in source.__dict__.items():
                if k == '_frozen':
                    continue
                setattr(self, k, v)

            self.__dict__['_frozen'] = True
            for i in self.extra:
                i.set_frozen()

        def __setattr__(self, k, v):
            if self._frozen:
                # This should really be an error, but we have a number of
                # drivers that make modifications during set_memory(). So this
                # just has to be a warning for now. Later it could turn into
                # a TypeError.
                caller = inspect.getframeinfo(inspect.stack()[1][0])
                LOG.warning(
                    '%s@%i: Illegal set on attribute %s - Fix this driver!' % (
                        caller.filename, caller.lineno, k))
            super().__setattr__(k, v)

        def dupe(self):
            m = Memory()
            m.clone(self)
            delattr(m, '_frozen')
            return m

    return _FrozenMemory(source)


class MemoryMapping(object):
    """Base class for a memory mapping"""

    def __init__(self, model, index, name):
        self._model = model
        self._index = index
        self._name = name

    def __str__(self):
        return self.get_name()

    def __repr__(self):
        return "%s-%s" % (self.__class__.__name__, self._index)

    def get_name(self):
        """Returns the mapping name"""
        return self._name

    def get_index(self):
        """Returns the immutable index (string or int)"""
        return self._index

    def __eq__(self, other):
        return self.get_index() == other.get_index()


class MappingModel(object):
    """Base class for a memory mapping model"""

    def __init__(self, radio, name):
        self._radio = radio
        self._name = name

    def get_name(self):
        return self._name

    def get_num_mappings(self):
        """Returns the number of mappings in the model (should be
        callable without consulting the radio"""
        raise NotImplementedError()

    def get_mappings(self):
        """Return a list of mappings"""
        raise NotImplementedError()

    def add_memory_to_mapping(self, memory, mapping):
        """Add @memory to @mapping."""
        raise NotImplementedError()

    def remove_memory_from_mapping(self, memory, mapping):
        """Remove @memory from @mapping.
        Shall raise exception if @memory is not in @bank"""
        raise NotImplementedError()

    def get_mapping_memories(self, mapping):
        """Return a list of memories in @mapping"""
        raise NotImplementedError()

    def get_memory_mappings(self, memory):
        """Return a list of mappings that @memory is in"""
        raise NotImplementedError()


class Bank(MemoryMapping):
    """Base class for a radio's Bank"""


class NamedBank(Bank):
    """A bank that can have a name"""

    def set_name(self, name):
        """Changes the user-adjustable bank name"""
        self._name = name


class BankModel(MappingModel):
    """A bank model where one memory is in zero or one banks at any point"""

    def __init__(self, radio, name='Banks'):
        super(BankModel, self).__init__(radio, name)


class MappingModelIndexInterface:
    """Interface for mappings with index capabilities"""

    def get_index_bounds(self):
        """Returns a tuple (lo,hi) of the min and max mapping indices"""
        raise NotImplementedError()

    def get_memory_index(self, memory, mapping):
        """Returns the index of @memory in @mapping"""
        raise NotImplementedError()

    def set_memory_index(self, memory, mapping, index):
        """Sets the index of @memory in @mapping to @index"""
        raise NotImplementedError()

    def get_next_mapping_index(self, mapping):
        """Returns the next available mapping index in @mapping, or raises
        Exception if full"""
        raise NotImplementedError()


class MTOBankModel(BankModel):
    """A bank model where one memory can be in multiple banks at once """
    pass


def console_status(status):
    """Write a status object to the console"""
    import logging
    from chirp import logger
    if not logger.is_visible(logging.WARN):
        return
    import sys
    import os
    sys.stdout.write("\r%s" % status)
    if status.cur == status.max:
        sys.stdout.write(os.linesep)


class RadioPrompts:
    """Radio prompt strings"""
    info = None
    experimental = None
    pre_download = None
    pre_upload = None
    display_pre_upload_prompt_before_opening_port = True


def BOOLEAN(v):
    assert v in (True, False)


def LIST(v):
    assert hasattr(v, '__iter__')


def INT(min=0, max=None):
    def checkint(v):
        assert isinstance(v, int)
        assert v >= min
        if max is not None:
            assert v <= max

    return checkint


def STRING(v):
    assert isinstance(v, str)


def NTUPLE(size):
    def checktuple(v):
        assert len(v) == size

    return checktuple


def TONELIST(v):
    assert all(VALIDTONE(x) for x in v)


class RadioFeatures:
    """Radio Feature Flags"""
    _valid_map = {
        # General
        "has_bank_index":       BOOLEAN,
        "has_dtcs":             BOOLEAN,
        "has_rx_dtcs":          BOOLEAN,
        "has_dtcs_polarity":    BOOLEAN,
        "has_mode":             BOOLEAN,
        "has_offset":           BOOLEAN,
        "has_name":             BOOLEAN,
        "has_bank":             BOOLEAN,
        "has_bank_names":       BOOLEAN,
        "has_tuning_step":      BOOLEAN,
        "has_ctone":            BOOLEAN,
        "has_cross":            BOOLEAN,
        "has_infinite_number":  BOOLEAN,
        "has_nostep_tuning":    BOOLEAN,
        "has_comment":          BOOLEAN,
        "has_settings":         BOOLEAN,
        "has_variable_power":   BOOLEAN,

        # Attributes
        "valid_modes":          LIST,
        "valid_tmodes":         LIST,
        "valid_duplexes":       LIST,
        "valid_tuning_steps":   LIST,
        "valid_bands":          LIST,
        "valid_skips":          LIST,
        "valid_power_levels":   LIST,
        "valid_characters":     STRING,
        "valid_name_length":    INT(),
        "valid_cross_modes":    LIST,
        "valid_tones":          TONELIST,
        "valid_dtcs_pols":      LIST,
        "valid_dtcs_codes":     LIST,
        "valid_special_chans":  LIST,

        "has_sub_devices":      BOOLEAN,
        "memory_bounds":        NTUPLE(2),
        "can_odd_split":        BOOLEAN,
        "can_delete":           BOOLEAN,

        # D-STAR
        "requires_call_lists":  BOOLEAN,
        "has_implicit_calls":   BOOLEAN,
    }

    def __setattr__(self, name, val):
        if name.startswith("_"):
            self.__dict__[name] = val
            return
        elif name not in list(self._valid_map.keys()):
            raise ValueError("No such attribute `%s'" % name)

        try:
            self._valid_map[name](val)
        except AssertionError:
            raise ValueError('Invalid value %r for attribute %r' % (
                val, name))

        self.__dict__[name] = val

    def __getattr__(self, name):
        raise AttributeError("pylint is confused by RadioFeatures")

    def init(self, attribute, default, doc=None):
        """Initialize a feature flag @attribute with default value @default,
        and documentation string @doc"""
        self.__setattr__(attribute, default)
        self.__docs[attribute] = doc

    def get_doc(self, attribute):
        """Return the description of @attribute"""
        return self.__docs[attribute]

    def __init__(self):
        self.__docs = {}
        self.init("has_bank_index", False,
                  "Indicates that memories in a bank can be stored in " +
                  "an order other than in main memory")
        self.init("has_dtcs", True,
                  "Indicates that DTCS tone mode is available")
        self.init("has_rx_dtcs", False,
                  "Indicates that radio can use two different " +
                  "DTCS codes for rx and tx")
        self.init("has_dtcs_polarity", True,
                  "Indicates that the DTCS polarity can be changed")
        self.init("has_mode", True,
                  "Indicates that multiple emission modes are supported")
        self.init("has_offset", True,
                  "Indicates that the TX offset memory property is supported")
        self.init("has_name", True,
                  "Indicates that an alphanumeric memory name is supported")
        self.init("has_bank", True,
                  "Indicates that memories may be placed into banks")
        self.init("has_bank_names", False,
                  "Indicates that banks may be named")
        self.init("has_tuning_step", True,
                  "Indicates that memories store their tuning step")
        self.init("has_ctone", True,
                  "Indicates that the radio keeps separate tone frequencies " +
                  "for repeater and CTCSS operation")
        self.init("has_cross", False,
                  "Indicates that the radios supports different tone modes " +
                  "on transmit and receive")
        self.init("has_infinite_number", False,
                  "Indicates that the radio is not constrained in the " +
                  "number of memories that it can store")
        self.init("has_nostep_tuning", False,
                  "Indicates that the radio does not require a valid " +
                  "tuning step to store a frequency")
        self.init("has_comment", False,
                  "Indicates that the radio supports storing a comment " +
                  "with each memory")
        self.init("has_settings", False,
                  "Indicates that the radio supports general settings")
        self.init("has_variable_power", False,
                  "Indicates the radio supports any power level between the "
                  "min and max in valid_power_levels")

        self.init("valid_modes", list(MODES),
                  "Supported emission (or receive) modes")
        self.init("valid_tmodes", [],
                  "Supported tone squelch modes")
        self.init("valid_duplexes", ["", "+", "-"],
                  "Supported duplex modes")
        self.init("valid_tuning_steps", list(COMMON_TUNING_STEPS),
                  "Supported tuning steps")
        self.init("valid_bands", [],
                  "Supported frequency ranges")
        self.init("valid_skips", ["", "S"],
                  "Supported memory scan skip settings")
        self.init("valid_power_levels", [],
                  "Supported power levels")
        self.init("valid_characters", CHARSET_UPPER_NUMERIC,
                  "Supported characters for a memory's alphanumeric tag")
        self.init("valid_name_length", 6,
                  "The maximum number of characters in a memory's " +
                  "alphanumeric tag")
        self.init("valid_cross_modes", list(CROSS_MODES),
                  "Supported tone cross modes")
        self.init("valid_tones", list(TONES),
                  "Support Tones")
        self.init("valid_dtcs_pols", ["NN", "RN", "NR", "RR"],
                  "Supported DTCS polarities")
        self.init("valid_dtcs_codes", list(DTCS_CODES),
                  "Supported DTCS codes")
        self.init("valid_special_chans", [],
                  "Supported special channel names")

        self.init("has_sub_devices", False,
                  "Indicates that the radio behaves as two semi-independent " +
                  "devices")
        self.init("memory_bounds", (0, 1),
                  "The minimum and maximum channel numbers")
        self.init("can_odd_split", False,
                  "Indicates that the radio can store an independent " +
                  "transmit frequency")
        self.init("can_delete", True,
                  "Indicates that the radio can delete memories")
        self.init("requires_call_lists", True,
                  "[D-STAR] Indicates that the radio requires all callsigns " +
                  "to be in the master list and cannot be stored " +
                  "arbitrarily in each memory channel")
        self.init("has_implicit_calls", False,
                  "[D-STAR] Indicates that the radio has an implied " +
                  "callsign at the beginning of the master URCALL list")

    def is_a_feature(self, name):
        """Returns True if @name is a valid feature flag name"""
        return name in list(self._valid_map.keys())

    def __getitem__(self, name):
        return self.__dict__[name]

    def validate_memory(self, mem):
        """Return a list of warnings and errors that will be encountered
        if trying to set @mem on the current radio"""
        msgs = []

        lo, hi = self.memory_bounds
        if not self.has_infinite_number and \
                (mem.number < lo or mem.number > hi) and \
                mem.extd_number not in self.valid_special_chans:
            msg = ValidationWarning("Location %i is out of range" % mem.number)
            msgs.append(msg)

        if (self.valid_modes and
                mem.mode not in self.valid_modes and
                'mode' not in mem.immutable and
                mem.mode != "Auto"):
            msg = ValidationError("Mode %s not supported" % mem.mode)
            msgs.append(msg)

        if self.valid_tmodes and mem.tmode not in self.valid_tmodes:
            msg = ValidationError("Tone mode %s not supported" % mem.tmode)
            msgs.append(msg)
        else:
            if mem.tmode == "Cross":
                if self.valid_cross_modes and \
                        mem.cross_mode not in self.valid_cross_modes:
                    msg = ValidationError("Cross tone mode %s not supported" %
                                          mem.cross_mode)
                    msgs.append(msg)

        if self.valid_tones and mem.rtone not in self.valid_tones:
            msg = ValidationError("Tone %.1f not supported" % mem.rtone)
            msgs.append(msg)
        if self.valid_tones and mem.ctone not in self.valid_tones:
            msg = ValidationError("Tone %.1f not supported" % mem.ctone)
            msgs.append(msg)

        if self.has_dtcs_polarity and \
                mem.dtcs_polarity not in self.valid_dtcs_pols:
            msg = ValidationError("DTCS Polarity %s not supported" %
                                  mem.dtcs_polarity)
            msgs.append(msg)

        if self.valid_dtcs_codes and \
                mem.dtcs not in self.valid_dtcs_codes:
            msg = ValidationError("DTCS Code %03i not supported" % mem.dtcs)
            msgs.append(msg)
        if self.valid_dtcs_codes and \
                mem.rx_dtcs not in self.valid_dtcs_codes:
            msg = ValidationError("DTCS Code %03i not supported" % mem.rx_dtcs)
            msgs.append(msg)

        if self.valid_duplexes and mem.duplex not in self.valid_duplexes:
            msg = ValidationError("Duplex %s not supported" % mem.duplex)
            msgs.append(msg)

        ts = mem.tuning_step
        if self.valid_tuning_steps and ts not in self.valid_tuning_steps and \
                not self.has_nostep_tuning:
            msg = ValidationError("Tuning step %.2f not supported" % ts)
            msgs.append(msg)

        if self.valid_bands:
            valid = False
            for lo, hi in self.valid_bands:
                if lo <= mem.freq < hi:
                    valid = True
                    break
            if not valid:
                msg = ValidationError(
                    ("Frequency {freq} is out "
                     "of supported range").format(freq=format_freq(mem.freq)))
                msgs.append(msg)

        if self.valid_bands and \
                self.valid_duplexes and \
                mem.duplex in ["split", "-", "+"]:
            if mem.duplex == "split":
                freq = mem.offset
            elif mem.duplex == "-":
                freq = mem.freq - mem.offset
            elif mem.duplex == "+":
                freq = mem.freq + mem.offset
            valid = False
            for lo, hi in self.valid_bands:
                if lo <= freq < hi:
                    valid = True
                    break
            if not valid:
                msg = ValidationError(
                    ("Tx freq {freq} is out "
                     "of supported range").format(freq=format_freq(freq)))
                msgs.append(msg)

        if mem.power and self.valid_power_levels:
            if self.has_variable_power:
                if (mem.power < min(self.valid_power_levels) or
                        mem.power > max(self.valid_power_levels)):
                    msg = ValidationWarning(
                        "Power level %s is out of radio's range" % mem.power)
                    msgs.append(msg)
            else:
                if mem.power not in self.valid_power_levels:
                    msg = ValidationWarning(
                        "Power level %s not supported" % mem.power)
                    msgs.append(msg)

        if self.valid_tuning_steps and not self.has_nostep_tuning:
            try:
                required_step(mem.freq, self.valid_tuning_steps)
            except errors.InvalidDataError as e:
                msgs.append(ValidationError(e))

        if self.valid_characters:
            for char in mem.name:
                if char not in self.valid_characters:
                    msgs.append(ValidationWarning("Name character " +
                                                  "`%s'" % char +
                                                  " not supported"))
                    break

        return msgs


class ValidationMessage(str):
    """Base class for Validation Errors and Warnings"""
    pass


class ValidationWarning(ValidationMessage):
    """A non-fatal warning during memory validation"""
    pass


class ValidationError(ValidationMessage):
    """A fatal error during memory validation"""
    pass


def split_validation_msgs(msgs):
    """Split a list of msgs into warnings,errors"""
    return ([x for x in msgs if isinstance(x, ValidationWarning)],
            [x for x in msgs if isinstance(x, ValidationError)])


class Alias(object):
    VENDOR = "Unknown"
    MODEL = "Unknown"
    VARIANT = ""


class Radio(Alias):
    """Base class for all Radio drivers"""
    BAUD_RATE = 9600
    # Whether or not we should use RTS/CTS flow control
    HARDWARE_FLOW = False
    # Whether or not we should assert DTR when opening the serial port
    WANTS_DTR = True
    # Whether or not we should assert RTS when opening the serial port
    WANTS_RTS = True
    ALIASES = []
    NEEDS_COMPAT_SERIAL = True
    FORMATS = []

    def status_fn(self, status):
        """Deliver @status to the UI"""
        console_status(status)

    def __init__(self, pipe):
        self.errors = []
        self.pipe = pipe

    def get_features(self):
        """Return a RadioFeatures object for this radio"""
        return RadioFeatures()

    @classmethod
    def get_name(cls):
        """Return a printable name for this radio"""
        return "%s %s" % (cls.VENDOR, cls.MODEL)

    @classmethod
    def get_prompts(cls):
        """Return a set of strings for use in prompts"""
        return RadioPrompts()

    def set_pipe(self, pipe):
        """Set the serial object to be used for communications"""
        self.pipe = pipe

    def get_memory(self, number):
        """Return a Memory object for the memory at location @number

        Constructs and returns a generic Memory object for the given location
        in the radio's memory. The memory should accurately represent what is
        actually stored in the radio as closely as possible. If the radio
        does not support changing some attributes of the location in question,
        the Memory.immutable list should be set appropriately.

        NB: No changes to the radio's memory should occur as a result of
        calling get_memory().
        """
        pass

    def erase_memory(self, number):
        """Erase memory at location @number"""
        mem = Memory()
        mem.number = number
        mem.empty = True
        self.set_memory(mem)

    def get_memories(self, lo=None, hi=None):
        """Get all the memories between @lo and @hi"""
        pass

    def set_memory(self, memory):
        """Set the memory object @memory

        This method should copy generic attributes from @memory to the
        radio's memory. It should not modify @memory and it should reproduce
        the generic attributes on @memory in the radio's memory as faithfully
        as the radio allows. Attributes that can't be copied exactly should
        be warned in validate_memory() with ValidationWarnings if a
        substitution will be made, or ValidationError if truly incompatible.
        In the latter case, set_memory() will not be called.
        """
        pass

    def get_mapping_models(self):
        """Returns a list of MappingModel objects (or an empty list)"""
        if hasattr(self, "get_bank_model"):
            # FIXME: Backwards compatibility for old bank models
            bank_model = self.get_bank_model()
            if bank_model:
                return [bank_model]
        return []

    def get_raw_memory(self, number):
        """Return a raw string describing the memory at @number"""
        pass

    def filter_name(self, name):
        """Filter @name to just the length and characters supported"""
        rf = self.get_features()
        if rf.valid_characters == rf.valid_characters.upper():
            # Radio only supports uppercase, so help out here
            name = name.upper()
        return "".join([x for x in name[:rf.valid_name_length]
                        if x in rf.valid_characters])

    def get_sub_devices(self):
        """Return a list of sub-device Radio objects, if
        RadioFeatures.has_sub_devices is True"""
        return []

    def validate_memory(self, mem):
        """Return a list of warnings and errors that will be encountered
        if trying to set @mem on the current radio"""
        rf = self.get_features()
        return rf.validate_memory(mem)

    def get_settings(self):
        """Returns a RadioSettings list containing one or more
        RadioSettingGroup or RadioSetting objects. These represent general
        setting knobs and dials that can be adjusted on the radio. If this
        function is implemented, the has_settings RadioFeatures flag should
        be True and set_settings() must be implemented as well."""
        pass

    def set_settings(self, settings):
        """Accepts the top-level RadioSettingGroup returned from get_settings()
        and adjusts the values in the radio accordingly. This function expects
        the entire RadioSettingGroup hierarchy returned from get_settings().
        If this function is implemented, the has_settings RadioFeatures flag
        should be True and get_settings() must be implemented as well."""
        pass

    @classmethod
    def supports_format(cls, fmt):
        """Returns true if file format @fmt is supported by this radio.

        This really should not be overridden by implementations
        without a good reason (like excluding one).
        """
        return fmt in cls.FORMATS

    def check_set_memory_immutable_policy(self, existing, new):
        """Checks whether or not a new memory will violate policy.

        Some radios have complex requirements for which fields of which
        memories can be modified at any given point. For the most part, radios
        require certain physical memory slots to have immutable fields
        (such as labels on call channels), and this default implementation
        checks that policy. However, other radios have more fine-grained
        rules in order to comply with FCC type acceptance, which requires
        overriding this behavior.

        ** This should almost never be overridden in your driver.

        ** This must not communicate with the radio, if implemented on a live-
           mode driver.
        """
        for field in existing.immutable:
            if getattr(existing, field) != getattr(new, field):
                raise ImmutableValueError(
                    'Field %s is not mutable on this memory' % field)


class ExternalMemoryProperties:
    """A mixin class that provides external memory property support.

    This is for use by drivers that have some way of storing additional
    memory properties externally (i.e. in metadata or a separate region)
    that cannot be loaded/updated during get_memory()/set_memory().
    Implementing this is much less ideal than supporting those properties
    directly, so this should only be used when absolutely necessary.
    """

    def get_memory_extra(self, memory):
        """Update @memory with extra fields.

        This is called after get_memory() and is passed the result
        for augmentation from external storage.
        """
        return memory

    def set_memory_extra(self, memory):
        """Update external storage of properties from @memory.

        This is called after set_memory() with the same memory object
        to record additional properties in external storage.
        """
        pass

    def erase_memory_extra(self, number):
        """Erase external storage for @memory.

        This is called after erase_memory() to clear external storage
        for memory @number.
        """
        pass

    def link_device_metadata(self, devices):
        """Link sub-device metadata with parent.

        This is called after get_sub_devices() to make sure that the
        sub-device instances use a reference into the main radio's
        metadata. In most cases, this does not need to be overridden.
        """
        # Link metadata of the sub-devices into the main by variant name
        # so that they are both included in a later save of the parent.
        for sub in devices:
            self._metadata.setdefault(sub.VARIANT, {})
            sub._metadata = self._metadata[sub.VARIANT]


class FileBackedRadio(Radio):
    """A file-backed radio stores its data in a file"""
    FILE_EXTENSION = 'dat'

    def save(self, filename):
        """Save the radio's memory map to @filename"""
        pass

    def load(self, filename):
        """Load the radio's memory map object from @filename"""
        pass


class DetectableInterface:
    DETECTED_MODELS = None

    @classmethod
    def detect_from_serial(cls, pipe):
        """Communicate with the radio via serial to determine proper class

        Returns an in implementation of CloneModeRadio if detected, or raises
        RadioError if not. If NotImplemented is raised, we assume that no
        detection is possible or necessary.
        """
        assert cls.DETECTED_MODELS is None, (
            'Class has detected models but no detect_from_serial() '
            'implementation')
        raise NotImplementedError()

    @classmethod
    def detected_models(cls):
        return list(cls.DETECTED_MODELS or [])


class CloneModeRadio(FileBackedRadio, ExternalMemoryProperties,
                     DetectableInterface):
    """A clone-mode radio does a full memory dump in and out and we store
    an image of the radio into an image file"""
    FILE_EXTENSION = "img"
    MAGIC = b'\x00\xffchirp\xeeimg\x00\x01'

    _memsize = 0

    def __init__(self, pipe):
        self.errors = []
        self._mmap = None
        self._memobj = None
        self._metadata = {}

        if isinstance(pipe, str):
            self.pipe = None
            self.load_mmap(pipe)
        elif isinstance(pipe, memmap.MemoryMapBytes):
            self.pipe = None
            self._mmap = pipe
            self.process_mmap()
        else:
            FileBackedRadio.__init__(self, pipe)

    def get_memsize(self):
        """Return the radio's memory size"""
        return self._memsize

    @classmethod
    def match_model(cls, filedata, filename):
        """Given contents of a stored file (@filedata), return True if
        this radio driver handles the represented model"""

        # Unless the radio driver does something smarter, claim
        # support if the data is the same size as our memory.
        # Ideally, each radio would perform an intelligent analysis to
        # make this determination to avoid model conflicts with
        # memories of the same size.
        return cls._memsize and len(filedata) == cls._memsize

    def sync_in(self):
        """Initiate a radio-to-PC clone operation"""
        pass

    def sync_out(self):
        """Initiate a PC-to-radio clone operation"""
        pass

    def save(self, filename):
        """Save the radio's memory map to @filename"""
        self.save_mmap(filename)

    def load(self, filename):
        """Load the radio's memory map object from @filename"""
        self.load_mmap(filename)

    def process_mmap(self):
        """Process a newly-loaded or downloaded memory map"""
        pass

    @classmethod
    def _strip_metadata(cls, raw_data):
        try:
            idx = raw_data.index(cls.MAGIC)
        except ValueError:
            LOG.debug('Image data has no metadata blob')
            return raw_data, {}

        # Find the beginning of the base64 blob
        raw_metadata = raw_data[idx + len(cls.MAGIC):]
        metadata = {}
        try:
            metadata = json.loads(base64.b64decode(raw_metadata).decode())
        except ValueError as e:
            LOG.error('Failed to parse decoded metadata blob: %s' % e)
        except TypeError as e:
            LOG.error('Failed to decode metadata blob: %s' % e)

        if metadata:
            LOG.debug('Loaded metadata: %s' % metadata)

        return raw_data[:idx], metadata

    def _make_metadata(self):
        # Always generate these directly from our in-memory state
        base = {
            'rclass': self.__class__.__name__,
            'vendor': self.VENDOR,
            'model': self.MODEL,
            'variant': self.VARIANT,
            'chirp_version': CHIRP_VERSION,
        }

        # Any other properties take a back seat to the above
        extra = {k: v for k, v in self._metadata.items() if k not in base}
        extra.update(base)

        return base64.b64encode(json.dumps(extra).encode())

    def load_mmap(self, filename):
        """Load the radio's memory map from @filename"""
        mapfile = open(filename, "rb")
        data = mapfile.read()
        if self.MAGIC in data:
            data, self._metadata = self._strip_metadata(data)
            if ('chirp_version' in self._metadata and
                    is_version_newer(self._metadata.get('chirp_version'))):
                LOG.warning('Image is from version %s but we are %s' % (
                    self._metadata.get('chirp_version'), CHIRP_VERSION))
        if self.NEEDS_COMPAT_SERIAL:
            self._mmap = memmap.MemoryMap(data)
        else:
            self._mmap = memmap.MemoryMapBytes(bytes(data))
        mapfile.close()
        self.process_mmap()

    def save_mmap(self, filename):
        """
        try to open a file and write to it
        If IOError raise a File Access Error Exception
        """
        try:
            mapfile = open(filename, "wb")
            mapfile.write(self._mmap.get_byte_compatible().get_packed())
            if filename.lower().endswith(".img"):
                mapfile.write(self.MAGIC)
                mapfile.write(self._make_metadata())
            mapfile.close()
        except IOError:
            raise Exception("File Access Error")

    def get_mmap(self):
        """Return the radio's memory map object"""
        return self._mmap

    @property
    def metadata(self):
        return dict(self._metadata)

    @metadata.setter
    def metadata(self, values):
        self._metadata.update(values)

    def get_memory_extra(self, memory):
        rf = self.get_features()
        if not rf.has_comment and isinstance(memory.number, int):
            self._metadata.setdefault('mem_extra', {})
            memory.comment = self._metadata['mem_extra'].get(
                '%04i_comment' % memory.number, '')
        return memory

    def set_memory_extra(self, memory):
        rf = self.get_features()
        if not rf.has_comment and isinstance(memory.number, int):
            self._metadata.setdefault('mem_extra', {})
            key = '%04i_comment' % memory.number
            if not memory.comment:
                self._metadata['mem_extra'].pop(key, None)
            else:
                self._metadata['mem_extra'][key] = memory.comment

    def erase_memory_extra(self, number):
        rf = self.get_features()
        if not rf.has_comment and isinstance(number, int):
            self._metadata.setdefault('mem_extra', {})
            self._metadata['mem_extra'].pop('%04i_comment' % number, None)


class LiveRadio(Radio, DetectableInterface):
    """Base class for all Live-Mode radios"""
    pass


class NetworkSourceRadio(Radio):
    """Base class for all radios based on a network source"""

    def do_fetch(self):
        """Fetch the source data from the network"""
        pass


class IcomDstarSupport:
    """Base interface for radios supporting Icom's D-STAR technology"""
    MYCALL_LIMIT = (1, 1)
    URCALL_LIMIT = (1, 1)
    RPTCALL_LIMIT = (1, 1)

    def get_urcall_list(self):
        """Return a list of URCALL callsigns"""
        return []

    def get_repeater_call_list(self):
        """Return a list of RPTCALL callsigns"""
        return []

    def get_mycall_list(self):
        """Return a list of MYCALL callsigns"""
        return []

    def set_urcall_list(self, calls):
        """Set the URCALL callsign list"""
        pass

    def set_repeater_call_list(self, calls):
        """Set the RPTCALL callsign list"""
        pass

    def set_mycall_list(self, calls):
        """Set the MYCALL callsign list"""
        pass


class ExperimentalRadio:
    """Interface for experimental radios"""
    @classmethod
    def get_experimental_warning(cls):
        return ("This radio's driver is marked as experimental and may " +
                "be unstable or unsafe to use.")


class Status:
    """Clone status object for conveying clone progress to the UI"""
    name = "Job"
    msg = "Unknown"
    max = 100
    cur = 0

    def __str__(self):
        try:
            pct = (self.cur / float(self.max)) * 100
            nticks = int(pct) // 10
            ticks = "=" * nticks
        except ValueError:
            pct = 0.0
            ticks = "?" * 10

        return "|%-10s| %2.1f%% %s" % (ticks, pct, self.msg)


def is_fractional_step(freq):
    """Returns True if @freq requires a 12.5 kHz or 6.25 kHz step"""
    return not is_5_0(freq) and (is_12_5(freq) or is_6_25(freq))


def is_5_0(freq):
    """Returns True if @freq is reachable by a 5 kHz step"""
    return (freq % 5000) == 0


def is_10_0(freq):
    """Returns True if @freq is reachable by a 10 kHz step"""
    return (freq % 10000) == 0


def is_12_5(freq):
    """Returns True if @freq is reachable by a 12.5 kHz step"""
    return (freq % 12500) == 0


def is_6_25(freq):
    """Returns True if @freq is reachable by a 6.25 kHz step"""
    return (freq % 6250) == 0


def is_2_5(freq):
    """Returns True if @freq is reachable by a 2.5 kHz step"""
    return (freq % 2500) == 0


def is_8_33(freq):
    """Returns True if @freq is reachable by a 8.33 kHz step"""
    return (freq % 25000) in [0, 8330, 16660]


def is_1_0(freq):
    """Returns True if @freq is reachable by a 1.0 kHz step"""
    return (freq % 1000) == 0


def is_0_5(freq):
    """Returns True if @freq is reachable by a 0.5 kHz step"""
    return (freq % 500) == 0


def make_is(stephz):
    def validator(freq):
        return freq % stephz == 0
    return validator


def required_step(freq, allowed=None):
    """Returns the simplest tuning step that is required to reach @freq"""
    if allowed is None:
        allowed = [5.0, 10.0, 12.5, 6.25, 2.5, 8.33]

    # These should be in order of most common to least common
    steps = {
        5.0: make_is(5000),
        10.0: make_is(10000),
        12.5: make_is(12500),
        6.25: make_is(6250),
        2.5: make_is(2500),
        1.0: make_is(1000),
        0.5: make_is(500),
        0.25: make_is(250),
        8.33: is_8_33,
    }

    for step, validate in steps.items():
        if step in allowed and validate(freq):
            return step

    raise errors.InvalidDataError("Unable to find a supported " +
                                  "tuning step for %s" % format_freq(freq))


def fix_rounded_step(freq):
    """Some radios imply the last bit of 12.5 kHz and 6.25 kHz step
    frequencies. Take the base @freq and return the corrected one"""
    try:
        required_step(freq)
        return freq
    except errors.InvalidDataError:
        pass

    try:
        required_step(freq + 500)
        return freq + 500
    except errors.InvalidDataError:
        pass

    try:
        required_step(freq + 250)
        return freq + 250
    except errors.InvalidDataError:
        pass

    try:
        required_step(freq + 750)
        return float(freq + 750)
    except errors.InvalidDataError:
        pass

    try:
        required_step(freq + 330)
        return float(freq + 330)
    except errors.InvalidDataError:
        pass

    try:
        required_step(freq + 660)
        return float(freq + 660)
    except errors.InvalidDataError:
        pass

    raise errors.InvalidDataError("Unable to correct rounded frequency " +
                                  format_freq(freq))


def _name(name, len, just_upper):
    """Justify @name to @len, optionally converting to all uppercase"""
    if just_upper:
        name = name.upper()
    return name.ljust(len)[:len]


def name6(name, just_upper=True):
    """6-char name"""
    return _name(name, 6, just_upper)


def name8(name, just_upper=False):
    """8-char name"""
    return _name(name, 8, just_upper)


def name16(name, just_upper=False):
    """16-char name"""
    return _name(name, 16, just_upper)


def to_GHz(val):
    """Convert @val in GHz to Hz"""
    return val * 1000000000


def to_MHz(val):
    """Convert @val in MHz to Hz"""
    return val * 1000000


def to_kHz(val):
    """Convert @val in kHz to Hz"""
    return val * 1000


def from_GHz(val):
    """Convert @val in Hz to GHz"""
    return val // 100000000


def from_MHz(val):
    """Convert @val in Hz to MHz"""
    return val // 100000


def from_kHz(val):
    """Convert @val in Hz to kHz"""
    return val // 100


def split_to_offset(mem, rxfreq, txfreq):
    """Set the freq, offset, and duplex fields of a memory based on
    a separate rx/tx frequency.
    """
    if abs(txfreq - rxfreq) > to_MHz(70):
        mem.freq = rxfreq
        mem.offset = txfreq
        mem.duplex = 'split'
    else:
        offset = txfreq - rxfreq
        if offset < 0:
            mem.duplex = '-'
        elif offset > 0:
            mem.duplex = '+'
        mem.offset = abs(offset)


def split_tone_decode(mem, txtone, rxtone):
    """
    Set tone mode and values on @mem based on txtone and rxtone specs like:
    None, None, None
    "Tone", 123.0, None
    "DTCS", 23, "N"
    """
    txmode, txval, txpol = txtone
    rxmode, rxval, rxpol = rxtone

    mem.dtcs_polarity = "%s%s" % (txpol or "N", rxpol or "N")

    if not txmode and not rxmode:
        # No tone
        return

    if txmode == "Tone" and not rxmode:
        mem.tmode = "Tone"
        mem.rtone = txval
        return

    if txmode == rxmode == "Tone" and txval == rxval:
        # TX and RX same tone -> TSQL
        mem.tmode = "TSQL"
        mem.ctone = txval
        return

    if txmode == rxmode == "DTCS" and txval == rxval:
        mem.tmode = "DTCS"
        mem.dtcs = txval
        return

    mem.tmode = "Cross"
    mem.cross_mode = "%s->%s" % (txmode or "", rxmode or "")

    if txmode == "Tone":
        mem.rtone = txval
    elif txmode == "DTCS":
        mem.dtcs = txval

    if rxmode == "Tone":
        mem.ctone = rxval
    elif rxmode == "DTCS":
        mem.rx_dtcs = rxval


def split_tone_encode(mem):
    """
    Returns TX, RX tone specs based on @mem like:
    None, None, None
    "Tone", 123.0, None
    "DTCS", 23, "N"
    """

    txmode = ''
    rxmode = ''
    txval = None
    rxval = None

    if mem.tmode == "Tone":
        txmode = "Tone"
        txval = mem.rtone
    elif mem.tmode == "TSQL":
        txmode = rxmode = "Tone"
        txval = rxval = mem.ctone
    elif mem.tmode == "DTCS":
        txmode = rxmode = "DTCS"
        txval = rxval = mem.dtcs
    elif mem.tmode == "Cross":
        txmode, rxmode = mem.cross_mode.split("->", 1)
        if txmode == "Tone":
            txval = mem.rtone
        elif txmode == "DTCS":
            txval = mem.dtcs
        if rxmode == "Tone":
            rxval = mem.ctone
        elif rxmode == "DTCS":
            rxval = mem.rx_dtcs

    if txmode == "DTCS":
        txpol = mem.dtcs_polarity[0]
    else:
        txpol = None
    if rxmode == "DTCS":
        rxpol = mem.dtcs_polarity[1]
    else:
        rxpol = None

    return ((txmode, txval, txpol),
            (rxmode, rxval, rxpol))


def sanitize_string(astring, validcharset=CHARSET_ASCII, replacechar='*'):
    myfilter = ''.join(
        [
            [replacechar, chr(x)][chr(x) in validcharset]
            for x in range(256)
        ])
    return astring.translate(myfilter)


def is_version_newer(version):
    """Return True if version is newer than ours"""

    def get_version(v):
        if v.startswith('daily-'):
            _, stamp = v.split('-', 1)
            ver = (int(stamp),)
        elif '.' in v:
            ver = tuple(int(p) for p in v.split('.'))
        else:
            ver = (0,)
        LOG.debug('Parsed version %r to %r' % (v, ver))
        return ver

    from chirp import CHIRP_VERSION

    try:
        version = get_version(version)
    except ValueError as e:
        LOG.error('Failed to parse version %r: %s' % (version, e))
        version = (0,)
    try:
        my_version = get_version(CHIRP_VERSION)
    except ValueError as e:
        LOG.error('Failed to parse my version %r: %s' % (CHIRP_VERSION, e))
        my_version = (0,)

    return version > my_version


def http_user_agent():
    ver = sys.version_info
    return 'chirp/%s (Python %i.%i.%i on %s)' % (
        CHIRP_VERSION,
        ver.major, ver.minor, ver.micro,
        sys.platform)


def urlretrieve(url, fn):
    """Grab an URL and save it in a specified file"""

    import urllib.request
    import urllib.error

    headers = {
        'User-Agent': http_user_agent(),
    }
    req = urllib.request.Request(url, headers=headers)
    resp = urllib.request.urlopen(req)
    with open(fn, 'wb') as f:
        f.write(resp.read())


def mem_from_text(text):
    m = Memory()
    freqs = re.findall(r'\b(\d{1,3}\.\d{2,6})\b', text)
    if not freqs:
        raise ValueError('Unable to find a frequency')
    m.freq = parse_freq(freqs[0])
    offset = re.search(r'([+-])\s*(\d\.\d{1,3}|\d)\b', text)
    duplex = re.search(r'\W([+-])\W', text[text.index(freqs[0]):])
    if len(freqs) > 1 and not offset:
        split_to_offset(m, m.freq, parse_freq(freqs[1]))
    else:
        if offset:
            m.duplex = offset.group(1)
            m.offset = parse_freq(offset.group(2))
        # Only look for the first +/- after the frequency, which would be
        # by far the most common arrangement
        if offset is None and duplex:
            m.duplex = duplex.group(1)
    tones = re.findall(r'\b(\d{2,3}\.\d|D\d{3})\b', text)
    if tones and len(tones) <= 2:
        txrx = []
        for val in tones:
            if '.' in val:
                mode = 'Tone'
                tone = float(val)
            elif 'D' in val:
                mode = 'DTCS'
                tone = int(val[1:])
            else:
                continue
            txrx.append((mode, tone, 'N'))
        if len(txrx) == 1:
            txrx.append(('', 88.5, 'N'))
        split_tone_decode(m, txrx[0], txrx[1])

    return m


def mem_to_text(mem):
    pieces = [format_freq(mem.freq)]
    if mem.duplex == 'split':
        pieces.append(format_freq(mem.offset))
    elif mem.duplex in ('-', '+'):
        pieces.append('%s%i.%3.3s' % (mem.duplex,
                                      mem.offset / 1000000,
                                      '%03i' % (mem.offset % 1000000)))
    txrx = split_tone_encode(mem)
    for mode, tone, pol in txrx:
        if mode == 'Tone':
            pieces.append('%.1f' % tone)
        elif mode == 'DTCS':
            pieces.append('D%03i' % tone)
    return '[%s]' % '/'.join(pieces)



# Copyright 2017 Dan Smith <dsmith@danplanet.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Rugged RH5R V2 radio management module"""

import struct
import logging

from chirp import chirp_common, bitwise, errors, directory, memmap


LOG = logging.getLogger(__name__)


def _identify(radio):
    try:
        radio.pipe.write("PGM2015")
        ack = radio.pipe.read(2)
        if ack != "\x06\x30":
            raise errors.RadioError("Radio did not ACK first command: %r" %
                                    ack)
    except:
        LOG.exception('')
        raise errors.RadioError("Unable to communicate with the radio")


def _download(radio):
    _identify(radio)
    data = []
    for i in range(0, 0x2000, 0x40):
        msg = struct.pack('>cHb', 'R', i, 0x40)
        radio.pipe.write(msg)
        block = radio.pipe.read(0x40 + 4)
        if len(block) != (0x40 + 4):
            raise errors.RadioError("Radio sent a short block (%02x/%02x)" % (
                len(block), 0x44))
        data += block[4:]

        if radio.status_fn:
            status = chirp_common.Status()
            status.cur = i
            status.max = 0x2000
            status.msg = "Cloning from radio"
            radio.status_fn(status)

    radio.pipe.write("E")
    data += 'PGM2015'

    return memmap.MemoryMap(data)


def _upload(radio):
    _identify(radio)
    for i in range(0, 0x2000, 0x40):
        msg = struct.pack('>cHb', 'W', i, 0x40)
        msg += radio._mmap[i:(i + 0x40)]
        radio.pipe.write(msg)
        ack = radio.pipe.read(1)
        if ack != '\x06':
            raise errors.RadioError('Radio did not ACK block %i (0x%04x)' % (
                i, i))

        if radio.status_fn:
            status = chirp_common.Status()
            status.cur = i
            status.max = 0x2000
            status.msg = "Cloning from radio"
            radio.status_fn(status)

    radio.pipe.write("E")


MEM_FORMAT = """
struct memory {
  bbcd rx_freq[4];
  bbcd tx_freq[4];
  lbcd rx_tone[2];
  lbcd tx_tone[2];

  u8 unknown10:5,
     highpower:1,
     unknown11:2;
  u8 unknown20:4,
     narrow:1,
     unknown21:3;
  u8 unknown31:1,
     scanadd:1,
     unknown32:6;
  u8 unknown4;
};

struct name {
  char name[7];
};

#seekto 0x0010;
struct memory channels[128];

#seekto 0x08C0;
struct name names[128];

#seekto 0x2020;
struct memory vfo1;
struct memory vfo2;
"""


POWER_LEVELS = [chirp_common.PowerLevel('Low', watts=1),
                chirp_common.PowerLevel('High', watts=5)]


class TYTTHUVF8_V2(chirp_common.CloneModeRadio):
    VENDOR = "TYT"
    MODEL = "TH-UVF8F"
    BAUD_RATE = 9600
    _FILEID = b'OEMOEM \\XFF'

    def get_features(self):
        rf = chirp_common.RadioFeatures()
        rf.memory_bounds = (1, 128)
        rf.has_bank = False
        rf.has_ctone = True
        rf.valid_tuning_steps = [5, 6.25, 10, 12.5]
        rf.has_tuning_step = False
        rf.has_cross = True
        rf.has_rx_dtcs = True
        rf.has_settings = False
        rf.can_odd_split = False
        rf.valid_duplexes = ['', '-', '+']
        rf.valid_tmodes = ["", "Tone", "TSQL", "DTCS", "Cross"]
        rf.valid_characters = chirp_common.CHARSET_UPPER_NUMERIC + "-"
        rf.valid_bands = [(136000000, 174000000),
                          (400000000, 480000000)]
        rf.valid_skips = ["", "S"]
        rf.valid_power_levels = POWER_LEVELS
        rf.valid_modes = ["FM", "NFM"]
        rf.valid_name_length = 7
        rf.valid_cross_modes = ["Tone->Tone", "Tone->DTCS", "DTCS->Tone",
                                "->Tone", "->DTCS", "DTCS->", "DTCS->DTCS"]
        return rf

    def sync_in(self):
        self._mmap = _download(self)
        self.process_mmap()

    def sync_out(self):
        _upload(self)

    @classmethod
    def match_model(cls, filedata, filename):
        return (filedata.endswith(b"PGM2015") and
                filedata[0x840:0x848] == cls._FILEID)

    def process_mmap(self):
        self._memobj = bitwise.parse(MEM_FORMAT, self._mmap)

    def get_raw_memory(self, number):
        return (repr(self._memobj.channels[number - 1]) +
                repr(self._memobj.names[number - 1]))

    def _get_memobjs(self, number):
        return (self._memobj.channels[number - 1],
                self._memobj.names[number - 1])

    def _decode_tone(self, toneval):
        pol = "N"
        rawval = (toneval[1].get_bits(0xFF) << 8) | toneval[0].get_bits(0xFF)

        if toneval[0].get_bits(0xFF) == 0xFF:
            mode = ""
            val = 0
        elif toneval[1].get_bits(0xC0) == 0xC0:
            mode = "DTCS"
            val = int("%x" % (rawval & 0x3FFF))
            pol = "R"
        elif toneval[1].get_bits(0x80):
            mode = "DTCS"
            val = int("%x" % (rawval & 0x3FFF))
        else:
            mode = "Tone"
            val = int(toneval) / 10.0

        return mode, val, pol

    def _encode_tone(self, _toneval, mode, val, pol):
        toneval = 0
        if mode == "Tone":
            toneval = int("%i" % (val * 10), 16)
        elif mode == "DTCS":
            toneval = int("%i" % val, 16)
            toneval |= 0x8000
            if pol == "R":
                toneval |= 0x4000
        else:
            toneval = 0xFFFF

        _toneval[0].set_raw(toneval & 0xFF)
        _toneval[1].set_raw((toneval >> 8) & 0xFF)

    def get_memory(self, number):
        _mem, _name = self._get_memobjs(number)

        mem = chirp_common.Memory()

        if isinstance(number, str):
            mem.number = SPECIALS[number]
            mem.extd_number = number
        else:
            mem.number = number

        if _mem.get_raw(asbytes=False).startswith("\xFF\xFF\xFF\xFF"):
            mem.empty = True
            return mem

        mem.freq = int(_mem.rx_freq) * 10
        offset = (int(_mem.tx_freq) - int(_mem.rx_freq)) * 10
        if not offset:
            mem.offset = 0
            mem.duplex = ''
        elif offset < 0:
            mem.offset = abs(offset)
            mem.duplex = '-'
        else:
            mem.offset = offset
            mem.duplex = '+'

        txmode, txval, txpol = self._decode_tone(_mem.tx_tone)
        rxmode, rxval, rxpol = self._decode_tone(_mem.rx_tone)

        chirp_common.split_tone_decode(mem,
                                       (txmode, txval, txpol),
                                       (rxmode, rxval, rxpol))

        mem.mode = 'NFM' if _mem.narrow else 'FM'
        mem.skip = '' if _mem.scanadd else 'S'
        mem.power = POWER_LEVELS[int(_mem.highpower)]
        mem.name = str(_name.name).rstrip('\xFF ')

        return mem

    def set_memory(self, mem):
        _mem, _name = self._get_memobjs(mem.number)
        if mem.empty:
            _mem.set_raw('\xFF' * 16)
            _name.set_raw('\xFF' * 7)
            return
        _mem.set_raw('\x00' * 16)

        _mem.rx_freq = mem.freq / 10
        if mem.duplex == '-':
            mult = -1
        elif not mem.duplex:
            mult = 0
        else:
            mult = 1
        _mem.tx_freq = (mem.freq + (mem.offset * mult)) / 10

        (txmode, txval, txpol), (rxmode, rxval, rxpol) = \
            chirp_common.split_tone_encode(mem)

        self._encode_tone(_mem.tx_tone, txmode, txval, txpol)
        self._encode_tone(_mem.rx_tone, rxmode, rxval, rxpol)

        _mem.narrow = mem.mode == 'NFM'
        _mem.scanadd = mem.skip != 'S'
        _mem.highpower = POWER_LEVELS.index(mem.power) if mem.power else 1
        _name.name = mem.name.rstrip(' ').ljust(7, '\xFF')


@directory.register
class RH5RV2(TYTTHUVF8_V2):
    VENDOR = "Rugged"
    MODEL = "RH5R-V2"
    _FILEID = b'RUGGED \xFF'
